"""
# Parameters for each query type

             1  2  3  4  5  6  7  8
───────────────────────────────────
   contract  ✓     ✓     ✓     ✓
     topic0  ✓  ✓        ✓  ✓
     topic1              ?  ?  ?  ?
     topic2              ?  ?  ?  ?
     topic3              ?  ?  ?  ?
start_block  ✓  ✓  ✓  ✓  ✓  ✓  ✓  ✓
  end_block  ✓  ✓  ✓  ✓  ✓  ✓  ✓  ✓
"""

import pytest

# from ctc.evm import event_utils_new as event_utils
from ctc.evm import event_utils


contract_address = '0x6b175474e89094c44da98b954eedeac495271d0f'
event_name = 'Transfer'
start_block = 14_200_000
end_block = 14_200_100

event_abi = {
    'anonymous': False,
    'inputs': [
        {
            'indexed': True,
            'internalType': 'address',
            'name': 'src',
            'type': 'address',
        },
        {
            'indexed': True,
            'internalType': 'address',
            'name': 'dst',
            'type': 'address',
        },
        {
            'indexed': False,
            'internalType': 'uint256',
            'name': 'wad',
            'type': 'uint256',
        },
    ],
    'name': 'Transfer',
    'type': 'event',
}


query_type_examples = [
    # 1
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'contract_address': contract_address,
            'event_abi': event_abi,
        },
        {
            'n_events': 62,
            'n_contracts': 1,
            'n_event_types': 1,
        },
    ),
    # 2
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'event_abi': event_abi,
        },
        {
            'n_events': 13645,
            'n_contracts': 1366,
            'n_event_types': 1,
        },
    ),
    # 3
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'contract_address': contract_address,
        },
        {
            'n_events': 69,
            'n_contracts': 1,
            'n_event_types': 2,
        },
    ),
    # 4
    (
        {
            'start_block': 14200001,
            'end_block': 14200001,
        },
        {
            'n_events': 25,
            'n_contracts': 10,
            'n_event_types': 5,
        },
    ),
    # 5
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'contract_address': contract_address,
            'event_abi': event_abi,
            'topic1': '0xa478c2975ab1ea89e8196811f51a7b7ade33eb11',
        },
        {
            'n_events': 5,
            'n_contracts': 1,
            'n_event_types': 1,
        },
    ),
    # 6
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'event_abi': event_abi,
            'topic1': '0xa478c2975ab1ea89e8196811f51a7b7ade33eb11',
        },
        {
            'n_events': 8,
            'n_contracts': 2,
            'n_event_types': 1,
        },
    ),
    # 7
    (
        {
            'start_block': start_block,
            'end_block': end_block,
            'contract_address': contract_address,
            'topic1': '0xa478c2975ab1ea89e8196811f51a7b7ade33eb11',
        },
        {
            'n_events': 5,
            'n_contracts': 1,
            'n_event_types': 1,
        },
    ),
]


@pytest.mark.parametrize('test', query_type_examples)
async def test_get_events_from_node(test):
    query, target_result = test
    actual_result = await event_utils.async_get_events(
        verbose=2,
        decode=False,
        context={'cache': False},
        **query,
    )

    assert len(actual_result) == target_result['n_events']
    assert (
        len(set(actual_result['contract_address']))
        == target_result['n_contracts']
    )
    assert (
        len(set(actual_result['event_hash'])) == target_result['n_event_types']
    )


@pytest.mark.parametrize('test', query_type_examples)
async def test_get_events_from_db(test):
    query, target_result = test
    actual_result = await event_utils.async_get_events(
        verbose=2,
        decode=False,
        context={'cache': True},
        **query,
    )

    assert len(actual_result) == target_result['n_events']
    assert (
        len(set(actual_result['contract_address']))
        == target_result['n_contracts']
    )
    assert (
        len(set(actual_result['event_hash'])) == target_result['n_event_types']
    )


@pytest.mark.parametrize('test', query_type_examples)
async def test_get_events_from_db_to_dict(test):
    query, target_result = test
    df = await event_utils.async_get_events(
        verbose=2,
        decode=False,
        context={'cache': True},
        **query,
    )
    actual_result = df.to_dicts()

    assert len(actual_result) == target_result['n_events']
    assert (
        len(set(event['contract_address'] for event in actual_result))
        == target_result['n_contracts']
    )
    assert (
        len(set(event['event_hash'] for event in actual_result))
        == target_result['n_event_types']
    )


@pytest.mark.parametrize('test', query_type_examples)
async def test_get_events_and_decode(test):
    query, target_result = test
    actual_result = await event_utils.async_get_events(
        verbose=2,
        decode=True,
        context={'cache': True},
        **query,
    )

    assert len(actual_result) == target_result['n_events']
    assert (
        len(set(actual_result['contract_address']))
        == target_result['n_contracts']
    )
    assert (
        len(set(actual_result['event_hash'])) == target_result['n_event_types']
    )


@pytest.mark.parametrize('test', query_type_examples)
async def test_get_events_and_decode_empty(test):
    query, target_result = test

    # change query to an empty block range
    target_result = dict(target_result)
    target_result['n_events'] = 0
    target_result['n_contracts'] = 0
    target_result['n_event_types'] = 0
    query = dict(query)
    query['start_block'] = 100
    query['end_block'] = 200

    actual_result = await event_utils.async_get_events(
        verbose=2,
        decode=True,
        context={'cache': True},
        **query,
    )

    assert len(actual_result) == target_result['n_events']
    assert (
        len(set(actual_result['contract_address']))
        == target_result['n_contracts']
    )
    assert (
        len(set(actual_result['event_hash'])) == target_result['n_event_types']
    )



from ctc import spec
from ctc.spec.typedefs import block_types
from ctc.spec.typedefs import db_types
from ctc.spec.typedefs import config_types
from ctc.spec.typedefs import transaction_types


def test_config_keys_match_types():
    assert set(spec.config_keys) == set(
        config_types.Config.__annotations__.keys()
    )
    assert set(spec.config_keys) == set(
        config_types.PartialConfig.__annotations__.keys()
    )
    assert set(spec.config_keys) == set(
        config_types.JsonConfig.__annotations__.keys()
    )


def test_block_keys_match_types():
    assert set(spec.rpc_block_keys) == set(
        block_types.RPCBlock.__annotations__.keys()
    )
    assert set(spec.db_block_keys) == set(
        block_types.DBBlock.__annotations__.keys()
    )


def test_transaction_keys_match_types():
    assert (
        set(spec.rpc_transaction_keys)
        == set(transaction_types.RPCTransaction.__annotations__.keys())
    )
    assert (
        set(spec.db_transaction_keys)
        == set(transaction_types.DBTransaction.__annotations__.keys())
    )


def test_schema_names_match_schema_types():
    assert set(spec.schema_names) == (
        set(db_types.AdminSchemaName.__args__)
        .union(set(db_types.GenericSchemaName.__args__))
        .union(set(db_types.NetworkSchemaName.__args__))
    )


def test_network_schema_names_match_schema_types():
    assert set(spec.network_schema_names) == set(
        db_types.NetworkSchemaName.__args__
    )




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.protocols import curve_utils
from .. import dex_class

if typing.TYPE_CHECKING:
    import tooltime


class CurveDEX(dex_class.DEX):
    """Curve DEX"""

    _pool_factories = {
        1: [
            '0xb9fc157394af804a3578134a6585c0dc9cc990d4',
            '0x90e00ace148ca3b23ac1bc8c240c2a7dd9c2d7f5',
            '0x0959158b6040d32d04c301a72cbfd6b39e21c9ae',
            '0x8f942c20d02befc377d41445793068908e2250d0',
            '0xf18056bbd320e96a48e3fbf8bc061322531aac99',
            '0xbabe61887f1de2713c6f97e567623453d3c79f67',
        ],
    }

    @classmethod
    async def async_get_new_pools(
        cls,
        *,
        factory: spec.Address,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:

        start_block, end_block = await evm.async_resolve_block_range(
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            allow_none=False,
            context=context,
        )

        start_block, end_block = await evm.async_block_numbers_to_int(
            [start_block, end_block], context=context
        )

        if factory == curve_utils.curve_deployer_eoa:
            unfiltered_pools = curve_utils.get_non_factory_pools()
            pools = []
            for pool, creation_block in unfiltered_pools.items():
                if start_block is not None and creation_block < start_block:
                    continue
                if end_block is not None and creation_block > end_block:
                    continue
                pools.append(pool)

        else:
            start_pool_count, end_pool_count = await rpc.async_batch_eth_call(
                to_address=factory,
                function_abi=curve_utils.function_abis['pool_count'],
                block_numbers=[start_block, end_block],
                context=context,
            )

            pools = await rpc.async_batch_eth_call(
                to_address=factory,
                function_abi=curve_utils.function_abis['pool_list'],
                function_parameter_list=[
                    [index] for index in range(start_pool_count, end_pool_count)
                ],
                context=context,
            )

            creation_blocks_coroutine = asyncio.create_task(
                evm.async_get_contracts_creation_blocks(
                    pools, context=context
                )
            )

        coroutines = [
            curve_utils.async_get_pool_tokens(pool, context=context)
            for pool in pools
        ]
        pools_tokens = await asyncio.gather(*coroutines)

        if factory == curve_utils.curve_deployer_eoa:
            creation_blocks = [unfiltered_pools[pool] for pool in pools]
        else:
            creation_blocks_fetched = await creation_blocks_coroutine
            if any(block is None for block in creation_blocks_fetched):
                raise Exception('could not find creation block for pool')
            else:
                creation_blocks = typing.cast(
                    typing.List[int], creation_blocks_fetched
                )

        dex_pools = []
        for pool, pool_tokens, creation_block in zip(
            pools, pools_tokens, creation_blocks
        ):

            if len(pool_tokens) < 4:
                pool_tokens = pool_tokens + [None] * (4 - len(pool_tokens))
            if creation_block is None:
                raise Exception('could not determine creation block of pool')

            asset0 = pool_tokens[0]
            asset1 = pool_tokens[1]
            asset2 = pool_tokens[2]
            asset3 = pool_tokens[3]

            dex_pool: spec.DexPool = {
                'address': pool,
                'factory': factory,
                'asset0': asset0,
                'asset1': asset1,
                'asset2': asset2,
                'asset3': asset3,
                'fee': None,
                'creation_block': creation_block,
                'additional_data': {},
            }
            dex_pools.append(dex_pool)

        return dex_pools

    @classmethod
    async def _async_get_pool_assets_from_node(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.Address]:
        return await curve_utils.async_get_pool_tokens(
            pool=pool,
            context=context,
        )

    @classmethod
    async def _async_get_pool_raw_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        include_timestamps: bool = False,
        verbose: bool = False,
        context: spec.Context = None,
    ) -> spec.RawDexTrades:

        import polars as pl

        # get data
        trades = await evm.async_get_events(
            contract_address=pool,
            event_abi=curve_utils.pool_event_abis['TokenExchange'],
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            verbose=verbose,
            include_timestamps=include_timestamps,
            context=context,
        )

        # check factory for whether TokenExchangeUnderlying is present
        contract_abi = await evm.async_get_contract_abi(pool, context=context)
        try:
            await evm.async_get_event_abi(
                contract_abi=contract_abi,
                event_name='TokenExchangeUnderlying',
                context=context,
            )
            has_token_exchange_underlying = True
        except LookupError:
            has_token_exchange_underlying = False

        if has_token_exchange_underlying:
            token_exchange_underlyings = await evm.async_get_events(
                pool,
                event_abi=curve_utils.pool_event_abis[
                    'TokenExchangeUnderlying'
                ],
                start_block=start_block,
                end_block=end_block,
                start_time=start_time,
                end_time=end_time,
                verbose=verbose,
                include_timestamps=include_timestamps,
                context=context,
            )
            trades = pl.concat([trades, token_exchange_underlyings])
            trades = trades.sort(['block_number', 'log_index'])

        # gather relevatn subset of data
        output: spec.RawDexTrades = {
            'block_number': trades['block_number'],
            'transaction_hash': trades['transaction_hash'],
            'recipient': trades['arg__buyer'],
            'sold_id': trades['arg__sold_id'],
            'bought_id': trades['arg__bought_id'],
            'sold_amount': trades['arg__tokens_sold'],
            'bought_amount': trades['arg__tokens_bought'],
        }

        if include_timestamps:
            output['timestamp'] = trades['timestamp']

        return output



from __future__ import annotations

import codecs
import typing

from ctc import spec


def get_binary_format(data: spec.GenericBinaryData) -> spec.BinaryFormat:
    """return binary format of input data"""
    if isinstance(data, bytes):
        return 'binary'
    elif isinstance(data, str):
        if data.startswith('0x'):
            return 'prefix_hex'
        else:
            return 'raw_hex'
    elif isinstance(data, int):
        return 'integer'
    else:
        raise Exception('could not detect format')


def get_binary_n_bytes(data: spec.GenericBinaryData) -> int:
    """return number of bytes in input data"""

    if isinstance(data, bytes):
        return len(data)
    elif isinstance(data, str):
        if len(data) % 2 != 0:
            raise Exception('hex data must have even number of characters')
        if data.startswith('0x'):
            return int(len(data) / 2) - 1
        else:
            return int(len(data) / 2)
    elif isinstance(data, int):
        # adapted from https://stackoverflow.com/a/30375198
        if data < 0:
            raise Exception('only positive integers allowed')
        return (data.bit_length() + 7) // 8
    else:
        raise Exception('unknown data type: ' + str(data))


#
# # binary data manipulation
#


def to_hex(
    data: spec.GenericBinaryData,
    *,
    prefix: bool = True,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> str:
    """convert hex, binary, or integer data to hex"""
    if prefix:
        output_format: typing.Literal['prefix_hex', 'raw_hex'] = 'prefix_hex'
    else:
        output_format = 'raw_hex'
    return binary_convert(
        data=data,
        output_format=output_format,
        n_bytes=n_bytes,
        keep_leading_0=keep_leading_0,
    )


def to_binary(
    data: spec.GenericBinaryData,
    *,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> bytes:
    """convert hex, binary, or integer data to binary"""
    return binary_convert(
        data=data,
        output_format='binary',
        n_bytes=n_bytes,
        keep_leading_0=keep_leading_0,
    )


def text_to_binary(
    text: str,
    output_format: typing.Optional[spec.BinaryFormat] = None,
) -> spec.GenericBinaryData:
    """convert text to binary data"""
    return binary_convert(text.encode(), output_format)


def binary_to_text(binary: spec.GenericBinaryData) -> str:
    """convert binary data to text"""
    return codecs.decode(binary_convert(binary, 'binary'))


@typing.overload
def binary_convert(
    data: spec.GenericBinaryData,
    output_format: typing.Literal['binary'],
    *,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> bytes:
    ...


@typing.overload
def binary_convert(
    data: spec.GenericBinaryData,
    output_format: typing.Literal['integer'],
    *,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> int:
    ...


@typing.overload
def binary_convert(
    data: spec.GenericBinaryData,
    output_format: typing.Optional[typing.Literal['prefix_hex', 'raw_hex']],
    *,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> str:
    ...


def binary_convert(
    data: spec.GenericBinaryData,
    output_format: typing.Optional[spec.BinaryFormat] = None,
    *,
    n_bytes: int | None = None,
    keep_leading_0: bool | None = None,
) -> spec.GenericBinaryData:
    """convert {hex str or bytes} into {hex str or bytes}

    function should not be used with general text data

    ## Data Types
    - 'prefix_hex': hex str with 0x prefix included
    - 'raw_hex': hex str without 0x prefix included
    - 'binary': bytes

    :data: binary data
    :output_format: str name of output format
    """

    if output_format is None:
        output_format = 'prefix_hex'

    if isinstance(data, str):
        if data.startswith('0x'):
            raw_data = data[2:]
        else:
            raw_data = data

        if n_bytes is not None and len(raw_data) / 2 != n_bytes:
            if len(raw_data) % 2 != 0:
                raise Exception('incomplete byte representation')
            input_bytes = int(len(raw_data) / 2)
            raw_data = '00' * (n_bytes - input_bytes) + raw_data

        if output_format == 'prefix_hex':
            return '0x' + raw_data
        elif output_format == 'raw_hex':
            return raw_data
        elif output_format == 'binary':
            if len(raw_data) % 2 == 1:
                raw_data = '0' + raw_data
            return bytes.fromhex(raw_data)
        elif output_format == 'integer':
            return int(data, 16)
        else:
            raise Exception('invalid output_format: ' + str(output_format))

    elif isinstance(data, bytes):

        if n_bytes is not None and len(data) != n_bytes:
            data = b'0' * (n_bytes - len(data)) + data

        if output_format == 'binary':
            return data
        elif output_format == 'prefix_hex':
            return '0x' + data.hex()
        elif output_format == 'raw_hex':
            return data.hex()
        elif output_format == 'integer':
            return int.from_bytes(data, 'big')
        else:
            raise Exception('invalid output_format: ' + str(output_format))

    elif isinstance(data, int):

        if data < 0:
            raise Exception('only positive integers allowed')

        if output_format == 'integer':

            return data

        else:

            if keep_leading_0 is None:
                keep_leading_0 = True
            if n_bytes is None:
                n_bytes = get_binary_n_bytes(data)
            as_bytes = data.to_bytes(n_bytes, 'big')

            if output_format == 'binary':
                return as_bytes
            elif output_format in ['prefix_hex', 'raw_hex']:
                as_hex = as_bytes.hex()
                if not keep_leading_0:
                    as_hex = as_hex.lstrip('0')
                    if data == 0:
                        as_hex = '0'
                if output_format == 'prefix_hex':
                    return '0x' + as_hex
                else:
                    return as_hex
            else:
                raise Exception('invalid output_format: ' + str(output_format))

    else:

        raise Exception('unknown input data format: ' + str(type(data)))


def binarize_fields(
    mapping: typing.Mapping[str, typing.Any],
    fields: typing.Sequence[str] | None = None,
    *,
    binary_format: spec.BinaryFormat = 'binary',
    allow_none: bool = True,
) -> typing.Mapping[str, typing.Any]:
    """convert fields of dict to specific binary format"""

    if fields is None:
        fields = list(mapping.keys())

    binarized: typing.MutableMapping[str, typing.Any] = {}
    for key, value in mapping.items():
        if key in fields:
            if value is not None:
                binarized[key] = binary_convert(value, binary_format)
            else:
                if allow_none:
                    binarized[key] = value
                else:
                    raise Exception('field has value None: ' + str(key))
        else:
            binarized[key] = value

    return binarized



from __future__ import annotations

import asyncio
import typing

import toolstr
import tooltime

from ctc import evm
from ctc import spec
from ctc import cli
from . import yearn_web_api
from . import yearn_strategies


async def async_search_single_vault_address(
    query: str,
    network: spec.NetworkReference,
) -> spec.Address:

    if evm.is_address_str(query):
        return query.lower()
    else:
        vault = await yearn_web_api.async_get_yearn_api_vault(
            query=query,
            network=network,
        )
        return vault['address']


async def async_print_vault_summary(
    query: str,
    network: spec.NetworkReference,
    *,
    verbose: bool = False,
    show_harvests: int | bool | None = None,
) -> None:

    import polars as pl

    obtained = tooltime.create_timestamp_iso_pretty()
    vault = await yearn_web_api.async_get_yearn_api_vault(
        query, network=network
    )

    styles = cli.get_cli_styles()

    toolstr.print_text_box(
        'Yearn Vault ' + vault['name'], style=styles['title']
    )

    print()
    toolstr.print_header('Vault Addresses', style=styles['title'])
    rows = [
        ['vault', vault['address'].lower()],
        ['token', vault['token']['address'].lower()],
    ]
    # if 'migration' in vault:
    #     rows.append(['migration', vault['migration']['address'].lower()])
    for strategy in vault['strategies']:
        rows.append([strategy['name'], strategy['address'].lower()])
    print()
    toolstr.print_table(
        rows,
        indent=4,
        border=styles['comment'],
        column_styles=[styles['option'], styles['metavar']],
    )

    inception_timestamp = await evm.async_get_block_timestamp(
        vault['inception']
    )
    age = tooltime.get_age(inception_timestamp, 'TimelengthPhrase')
    rough_age = ', '.join(age.split(', ')[:2])

    print()
    print()
    toolstr.print_header('Vault Metadata', style=styles['title'])
    print()
    rows = [
        ['name', vault['name']],
        ['display name', vault['display_name']],
        ['symbol', vault['symbol']],
        ['version', vault['version']],
        ['type', vault['type']],
        ['decimals', vault['decimals']],
        ['inception block', str(vault['inception'])],
        ['age', rough_age],
    ]
    if verbose:
        fees = vault['apy']['fees']
        rows.append(
            [
                'management fee',
                toolstr.format(fees['management'], percentage=True),
            ]
        )
        rows.append(
            [
                'performance fee',
                toolstr.format(fees['performance'], percentage=True),
            ]
        )
        if vault['apy']['type'] == 'crv':
            rows.append(
                ['keep CRV', toolstr.format(fees['keep_crv'], percentage=True)]
            )
            rows.append(
                [
                    'CVX keep CRV',
                    toolstr.format(fees['cvx_keep_crv'], percentage=True),
                ]
            )

    toolstr.print_table(
        rows,
        indent=4,
        border=styles['comment'],
        column_styles=[styles['option'], styles['description']],
    )

    if verbose:
        print()
        print()
        toolstr.print_header('Token Metadata', style=styles['title'])
        print()
        rows = [
            ['token name', vault['token']['name']],
            ['token display name', vault['token']['display_name']],
            ['token symbol', vault['token']['symbol']],
            ['token decimals', vault['token']['decimals']],
        ]
        toolstr.print_table(
            rows,
            indent=4,
            border=styles['comment'],
            column_styles=[styles['option'], styles['description']],
        )

    print()
    print()
    toolstr.print_header('Vault State', style=styles['title'])
    print()
    tvl = toolstr.format(
        vault['tvl']['tvl'],
        prefix='$',
        decimals=2,
        trailing_zeros=True,
        order_of_magnitude=True,
    )
    gross_apr = toolstr.format(
        vault['apy']['gross_apr'],
        percentage=True,
        decimals=3,
        trailing_zeros=True,
    )
    net_apy = toolstr.format(
        vault['apy']['net_apy'],
        percentage=True,
        decimals=3,
        trailing_zeros=True,
    )

    rows = [
        ['TVL', tvl],
        ['gross APR', gross_apr],
        ['net APY', net_apy],
    ]
    if verbose:
        rows.append(['APR type', vault['apy']['type']])
        rows.append(['is shutdown', vault['emergency_shutdown']])
        rows.append(
            ['last updated', tooltime.timestamp_to_iso_pretty(vault['updated'])]
        )
    toolstr.print_table(
        rows,
        indent=4,
        border=styles['comment'],
        column_styles=[styles['option'], styles['description']],
    )

    # display harvests data
    # TODO: this is very unoptimized, wait until after event db overhaul is complete
    if show_harvests not in [None, False]:

        print()
        print()
        toolstr.print_header('Recent Harvests', style=styles['title'])

        if isinstance(show_harvests, bool):
            n_harvests = 10
        elif isinstance(show_harvests, int):
            n_harvests = show_harvests
        else:
            raise Exception('unknown format for show_harvests parameter')

        strategies = [strategy['address'] for strategy in vault['strategies']]
        strategy_address_to_name = {
            strategy['address'].lower(): strategy['name']
            for strategy in vault['strategies']
        }
        try:
            coroutines = [
                yearn_strategies.async_get_harvests(strategy=strategy)
                for strategy in strategies
            ]

            results = await asyncio.gather(*coroutines)
            all_harvests = pl.concat(results)
            all_harvests = all_harvests.sort('block_number', 'log_index')
        except Exception:
            print()
            toolstr.print(
                'could not obtain harvests',
                style=styles['content'],
                indent=4,
            )
            all_harvests = pl.DataFrame()
        rows = []
        for harvest in all_harvests[-n_harvests:].to_dicts():
            age = tooltime.get_age(harvest['timestamp'], 'TimelengthPhrase')
            age = ', '.join(age.split(', ')[:1])
            row = [
                age,
                harvest['apr'],
                strategy_address_to_name[harvest['contract_address']],
            ]
            rows.append(row)
        labels = [
            'age',
            'apr',
            'strategy',
        ]
        print()
        toolstr.print_table(
            rows=rows,
            labels=labels,
            indent=4,
            column_formats={
                'apr': {
                    'percentage': True,
                    'trailing_zeros': True,
                    'decimals': 3,
                },
            },
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'age': styles['description'],
                'apr': styles['description'],
                'strategy': styles['metavar'],
            },
        )

    print()
    toolstr.print(
        'data from Yearn web API at ' + obtained,
        style=styles['comment'],
    )


async def async_print_vaults_summary(
    *,
    network: spec.NetworkReference,
    sort_by: str | None = None,
    min_tvl: int | float | None = None,
    min_apy: int | float | None = None,
    min_apr: int | float | None = None,
    n: int | None = 20,
    verbose: bool = False,
) -> None:

    if sort_by is None:
        sort_by = 'gross APR'

    api_vaults = await yearn_web_api.async_get_yearn_api_vaults(network=network)

    n_all_vaults = len(api_vaults)

    # filter data
    n_clipped = 0
    filtered_data = []
    for datum in api_vaults:
        tvl = datum['tvl']['tvl']
        if tvl is None:
            tvl = 0
        net_apy = datum['apy']['net_apy']
        gross_apr = datum['apy']['gross_apr']
        if (
            (min_tvl is not None and tvl < min_tvl)
            or (min_apy is not None and net_apy < min_apy)
            or (min_apr is not None and gross_apr < min_apr)
        ):
            n_clipped += 1
            continue
        filtered_data.append(datum)
    n_matching_vaults = len(filtered_data)
    data = filtered_data

    # build rows
    labels = [
        'name',
        'TVL',
        'net APY',
        'gross APR',
    ]
    if verbose:
        labels.append('address')
    total_tvl = 0
    rows = []
    for datum in data:

        tvl = datum['tvl']['tvl']
        net_apy = datum['apy']['net_apy']
        gross_apr = datum['apy']['gross_apr']

        if tvl is None:
            tvl = 0
        if tvl < 0.01:
            tvl = 0
        total_tvl += tvl

        row = [
            datum['name'],
            tvl,
            net_apy,
            gross_apr,
        ]
        if verbose:
            row.append(datum['address'].lower())
        rows.append(row)

    # sort rows
    sort_index = labels.index(sort_by)
    rows = sorted(rows, key=lambda row: row[sort_index], reverse=True)  # type: ignore

    # clip display rows
    if n is not None and len(rows) > n:
        rows = rows[:n]
        clip_row = ['...'] * len(labels)
        rows.append(clip_row)

    # add total row
    blank_row = [''] * len(labels)
    total_row: typing.List[typing.Any] = [''] * len(labels)
    total_row[labels.index('name')] = 'TOTAL'
    total_row[labels.index('TVL')] = total_tvl
    rows.append(blank_row)
    rows.append(total_row)

    # print outputs
    styles = cli.get_cli_styles()
    toolstr.print_text_box('Yearn Pool Summary', style=styles['title'])
    toolstr.print(
        'data obtained from https://api.yearn.finance', style=styles['comment']
    )
    if n_matching_vaults != n_all_vaults:
        print()
        toolstr.print(
            toolstr.add_style(str(n_matching_vaults), styles['description'])
            + ' / '
            + toolstr.add_style(str(n_all_vaults), styles['description'])
            + ' vaults match input filters'
        )
    print()
    if verbose:
        indent = None
    else:
        indent = 4
    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'name': styles['metavar'],
            'TVL': styles['description'],
            'net APY': styles['description'],
            'gross APR': styles['description'],
            'address': styles['metavar'],
        },
        indent=indent,
        column_formats={
            'TVL': {
                'order_of_magnitude': True,
                'trailing_zeros': True,
                'decimals': 2,
                'prefix': '$',
            },
            'net APY': {
                'percentage': True,
                'trailing_zeros': True,
                'decimals': 2,
            },
            'gross APR': {
                'percentage': True,
                'trailing_zeros': True,
                'decimals': 2,
            },
        },
        compact=verbose,
    )



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': hex_command,
        'help': 'convert text data to hex representation',
        'args': [
            {'name': 'text', 'help': 'ascii or other text data to convert'},
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'omit the "0x" prefix',
            },
        ],
        'examples': [
            'vitalik',
            'turducken',
        ],
    }


def hex_command(text: str, raw: bool) -> None:
    if raw:
        output = evm.text_to_binary(text, 'raw_hex')
    else:
        output = evm.text_to_binary(text, 'prefix_hex')
    print(output)



"""
see ctc.db.schemas.events.events_intake for db writing functions
"""
from __future__ import annotations

import typing

from ctc import spec
from . import event_node_utils
from . import event_query_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


async def _async_plan_event_query(
    *,
    contract_address: spec.Address | None,
    event_hash: bytes | None,
    topic1: bytes | None,
    topic2: bytes | None,
    topic3: bytes | None,
    start_block: int,
    end_block: int,
    context: spec.Context = None,
) -> spec.EventQueryPlan:
    """plan which parts of query should be dispatched to db vs node"""

    from ctc import config
    from ctc import db
    from ctc.toolbox import range_utils

    # parse query type
    query_type = event_query_utils._parse_event_query_type(
        contract_address=contract_address,
        event_hash=event_hash,
        topic1=topic1,
        topic2=topic2,
        topic3=topic3,
    )

    # get previous queries
    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='events', context=context
    )
    if not read_cache:
        db_ranges: typing.Sequence[typing.Sequence[int]] = []
    else:
        db_queries = await db.async_query_event_queries(
            query_type=query_type,
            contract_address=contract_address,
            event_hash=event_hash,
            topic1=topic1,
            topic2=topic2,
            topic3=topic3,
            context=context,
            start_block=start_block,
            end_block=end_block,
        )
        if db_queries is not None:
            db_ranges = []
            for db_query in db_queries:
                db_start_block = db_query['start_block']
                db_end_block = db_query['end_block']
                if db_start_block <= start_block:
                    db_start_block = start_block
                if db_end_block >= end_block:
                    db_end_block = end_block
                db_ranges.append((db_start_block, db_end_block))

        else:
            db_ranges = []

    # get overlap between current query range and previous query ranges
    db_ranges = range_utils.combine_overlapping_ranges(
        db_ranges,
        include_contiguous=True,
    )
    node_ranges = range_utils.get_range_gaps(
        start=start_block,
        end=end_block,
        subranges=db_ranges,
    )
    db_event_queries: typing.Sequence[spec.EventQuery] = [
        {
            'contract_address': contract_address,
            'event_hash': event_hash,
            'topic1': topic1,
            'topic2': topic2,
            'topic3': topic3,
            'start_block': db_start,
            'end_block': db_end,
        }
        for db_start, db_end in db_ranges
    ]

    # queries
    node_queries: typing.Sequence[spec.EventQuery] = [
        {
            'contract_address': contract_address,
            'event_hash': event_hash,
            'topic1': topic1,
            'topic2': topic2,
            'topic3': topic3,
            'start_block': node_start,
            'end_block': node_end,
        }
        for node_start, node_end in node_ranges
    ]

    return {
        'node': node_queries,
        'db': db_event_queries,
    }


async def _async_query_events_from_node_and_db(
    *,
    contract_address: spec.Address | None,
    event_hash: bytes | None,
    topic1: bytes | None,
    topic2: bytes | None,
    topic3: bytes | None,
    start_block: int,
    end_block: int,
    context: spec.Context,
    verbose: bool | int,
    columns_to_load: typing.Sequence[str],
    binary_output_format: Literal['binary', 'prefix_hex'] = 'binary',
    max_blocks_per_request: int = 2000,
) -> spec.DataFrame:

    import polars as pl
    from ctc import db
    from ctc.toolbox import pl_utils

    # check which portion of query is in db
    queries = await _async_plan_event_query(
        contract_address=contract_address,
        event_hash=event_hash,
        topic1=topic1,
        topic2=topic2,
        topic3=topic3,
        start_block=start_block,
        end_block=end_block,
        context=context,
    )

    if verbose >= 2:
        print('Query Plan:')
        print('- db queries:', len(queries['db']))
        print('- node queries:', len(queries['node']))

    results: typing.MutableMapping[int, spec.DataFrame] = {}

    # TODO: query from db and node in parallel

    # send db queries to db
    for query in queries['db']:
        db_result: spec.DataFrame | None = await db.async_query_events(
            context=context,
            binary_output_format=binary_output_format,
            columns=columns_to_load,
            output_format='polars',
            **query,
        )

        if db_result is None:
            continue

        # adjust types
        recast_bools = pl.col(pl.datatypes.Boolean).cast(pl.datatypes.Utf8)
        db_result = db_result.with_columns(db_result.select(recast_bools))
        if binary_output_format == 'prefix_hex':
            db_result = pl_utils.binary_columns_to_prefix_hex(db_result)

        results[query['start_block']] = db_result

    # send node queries to node
    for query in queries['node']:

        result = await event_node_utils._async_query_events_from_node(
            context=context,
            verbose=verbose,
            binary_output_format=binary_output_format,
            max_blocks_per_request=max_blocks_per_request,
            **query,
        )

        # omit unwanted columns
        for key in {
            'transaction_index',
            'log_index',
            'transaction_hash',
            'contract_address',
            'event_hash',
            'topic1',
            'topic2',
            'topic3',
            'unindexed',
        }:
            if key not in columns_to_load:
                result = result.drop(key)

        if binary_output_format == 'binary':
            result = pl_utils.prefix_hex_columns_to_binary(result)

        results[query['start_block']] = result

    sorted_results = [results[key] for key in sorted(results.keys())]
    events = pl.concat(sorted_results)

    return events




from __future__ import annotations

import toolcli

from .. import llama_tvls


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_protocol_command,
        'help': 'output data about a protocol tracked by Defi Llama',
        'args': [
            {
                'name': 'protocol',
                'help': 'protocol to display information about',
            },
            {
                'name': ['-v', '--verbose'],
                'help': 'display additional infromation',
                'action': 'store_true',
            },
        ],
        'examples': [
            'Balancer',
            'Balancer --verbose',
        ],
    }


async def async_llama_protocol_command(protocol: str, verbose: bool) -> None:
    await llama_tvls.async_print_historical_protocol_tvl(
        protocol=protocol,
        verbose=verbose,
    )



from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class DataSource(TypedDict, total=False):
        protocol: typing.Literal['UniswapV2', 'Chainlink']
        feed: typing.Union[str, spec.Address]
        composite_feed: typing.Union[str, spec.Address]
        invert: bool
        normalize: bool
        mode: typing.Literal['native', 'raw']




from __future__ import annotations

import typing

from .. import binary_utils
from ctc import spec


def encode_block_number(block: spec.BlockNumberReference) -> str:
    """encode block number as hex str"""

    standard_block = standardize_block_number(block)

    if isinstance(standard_block, str):
        return standard_block
    elif isinstance(standard_block, int):
        return binary_utils.to_hex(standard_block, keep_leading_0=False)
    else:
        raise Exception('could not encode block number')


#
# # singular block standardization
#


def standardize_block_number(
    block: typing.Optional[spec.BlockNumberReference],
) -> spec.StandardBlockNumber:
    """turn block into standard block number reference

    Examples of standard block numbers: 'latest' or 123456
    """

    if block is None:
        block = 'latest'

    if block in spec.block_number_names:
        if typing.TYPE_CHECKING:
            return typing.cast(spec.BlockNumberName, block)
        else:
            return block
    else:
        return raw_block_number_to_int(block)


def raw_block_number_to_int(block: spec.RawBlockNumber) -> int:
    """convert block number to int"""

    # python3.7 compatibility
    # supports_int = isinstance(block, typing.SupportsInt)
    supports_int = hasattr(block, '__int__')

    if supports_int:
        if isinstance(block, int):
            return block
        else:
            block_supports_int = typing.cast(typing.SupportsInt, block)
            as_int = int(round(block_supports_int))  # type: ignore
            if abs(as_int - int(block_supports_int)) > 0.0001:
                raise Exception('must specify integer blocks')
            return as_int
    elif isinstance(block, str):
        if block.startswith('0x'):
            try:
                return int(block, 16)
            except ValueError:
                pass
        elif 'e' in block:
            as_float = float(block)
            as_int = int(as_float)
            if abs(as_float - as_int) > 0.0001:
                raise Exception('must specify integer blocks')
            else:
                return as_int
        else:
            return int(block)

    raise Exception('unknown block number specification: ' + str(block))


#
# # plural block standardization
#


def standardize_block_numbers(
    blocks: typing.Iterable[spec.BlockNumberReference],
) -> list[spec.StandardBlockNumber]:
    """standardize an iterable of block number references"""
    return [standardize_block_number(block) for block in blocks]


def raw_block_numbers_to_ints(
    blocks: typing.Iterable[spec.RawBlockNumber],
) -> list[int]:
    """convert block numbers to integer"""
    return [raw_block_number_to_int(block) for block in blocks]



# isinstance(address, str) and len(address) > 4 and address.endswith('.eth'):
from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from ctc import spec

#     N = typing.TypeVar('N')


# def is_ens_name(name: typing.Any) -> bool:
#     return (
#         isinstance(name, str)
#         and len(name) > 4
#         and name.endswith('.eth')
#     )


async def async_resolve_address(
    name_or_address: str,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> str:
    """resolve alternate address formats (i.e. ENS) to standard address str"""

    if not (
        isinstance(name_or_address, str)
        and len(name_or_address) > 4
        and name_or_address.endswith('.eth')
    ):
        return name_or_address

    from ctc.protocols import ens_utils

    result = await ens_utils.async_resolve_name(
        name=name_or_address,
        context=context,
        block=block,
    )
    if result is not None:
        return result
    else:
        return name_or_address


async def async_resolve_addresses(
    names_or_addresses: typing.Sequence[str],
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[str]:
    """resolve alternate address formats (i.e. ENS) to standard address strs"""

    to_resolve: list[str] = []
    for item in names_or_addresses:
        if isinstance(item, str) and len(item) > 4 and item.endswith('.eth'):
            to_resolve.append(item)

    if len(to_resolve) == 0:
        return names_or_addresses

    else:

        from ctc.protocols import ens_utils

        results = await ens_utils.async_resolve_names(
            names=to_resolve,
            block=block,
            context=context,
        )
        resolved: typing.Mapping[str, str | None] = dict(
            zip(to_resolve, results)
        )

        output: list[str] = []
        for name_or_address in names_or_addresses:
            resolved_name_or_address = resolved.get(name_or_address)
            if resolved_name_or_address is not None:
                output.append(resolved_name_or_address)
            else:
                output.append(name_or_address)

        return output


async def async_resolve_address_by_block(
    name_or_address: str,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    context: spec.Context = None,
) -> typing.Sequence[str]:
    """return resolution history of address over multiple blocks"""

    import asyncio

    coroutines = [
        async_resolve_address(
            name_or_address=name_or_address,
            block=block,
            context=context,
        )
        for block in blocks
    ]

    results = await asyncio.gather(*coroutines)

    return [
        (result if result is not None else name_or_address)
        for result in results
    ]



from __future__ import annotations

import typing

import toolsql

from ctc import spec
from ... import schema_utils


async def async_upsert_block_timestamp(
    *,
    conn: toolsql.AsyncConnection,
    block_number: int,
    timestamp: int,
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row={'block_number': block_number, 'timestamp': timestamp},
        upsert=True,
    )


async def async_upsert_block_timestamps(
    *,
    conn: toolsql.AsyncConnection,
    block_timestamps: typing.Mapping[int, int] | None = None,
    blocks: typing.Sequence[spec.DBBlock] | None = None,
    context: spec.Context = None,
) -> None:

    if block_timestamps is None:
        if blocks is None:
            raise Exception('must specify block_timestamps or blocks')
        block_timestamps = {
            block['number']: block['timestamp'] for block in blocks
        }

    rows = [
        {'block_number': block_number, 'timestamp': timestamp}
        for block_number, timestamp in block_timestamps.items()
    ]
    table = schema_utils.get_table_schema('block_timestamps', context=context)
    await toolsql.async_insert(
        conn=conn,
        table=table,
        rows=rows,
        upsert=True,
    )


async def async_delete_block_timestamp(
    *,
    conn: toolsql.AsyncConnection,
    block_number: typing.Sequence[int],
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
    )


async def async_delete_block_timestamps(
    *,
    conn: toolsql.AsyncConnection,
    block_numbers: typing.Sequence[int],
    context: spec.Context = None,
) -> None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    await toolsql.async_delete(
        conn=conn,
        table=table,
        where_in={'block_number': block_numbers},
    )


#
# # do not export
#


async def async_select_block_timestamp(
    block_number: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    result = await toolsql.async_select(
        conn=conn,
        table=table,
        where_equals={'block_number': block_number},
        columns=['timestamp'],
        output_format='cell_or_none',
    )
    if result is not None and not isinstance(result, int):
        raise Exception('invalid db result')
    return result


async def async_select_block_timestamps(
    block_numbers: typing.Sequence[typing.SupportsInt],
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> list[int | None] | None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    block_numbers_int = [int(item) for item in block_numbers]

    results = await toolsql.async_select(
        conn=conn,
        table=table,
        where_in={'block_number': block_numbers_int},
    )

    if results is None:
        return results

    block_timestamps = {
        row['block_number']: row['timestamp'] for row in results
    }

    return [
        block_timestamps.get(block_number) for block_number in block_numbers
    ]


async def async_select_max_block_number(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    # try-catch because connectorx panics on empty rows
    try:
        result = await toolsql.async_select(
            conn=conn,
            table=table,
            columns=['MAX(block_number)'],
            output_format='cell_or_none',
        )
    except RuntimeError as e:
        if e.args == ('Cannot infer type from null for SQLite',):
            return None
        else:
            raise e

    if result is None:
        return None
    elif isinstance(result, int):
        return result
    else:
        raise Exception('invalid db result')


async def async_select_max_block_timestamp(
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> int | None:

    table = schema_utils.get_table_schema('block_timestamps', context=context)

    # try-catch because connectorx panics on empty rows
    try:
        result = await toolsql.async_select(
            conn=conn,
            table=table,
            columns=['MAX(timestamp)'],
            output_format='cell_or_none',
        )
    except RuntimeError as e:
        if e.args == ('Cannot infer type from null for SQLite',):
            return None
        else:
            raise e

    if result is None:
        return None
    elif isinstance(result, int):
        return result
    else:
        raise Exception('invalid db result')


async def async_select_timestamp_block_range(
    timestamp: int,
    *,
    conn: toolsql.AsyncConnection,
    context: spec.Context = None,
) -> tuple[int | None, int | None]:
    """return block range that must contain timestamp

    this function is used to confine a block-of-timestamp search
    """

    # convert numpy types to native python type
    if type(timestamp).__name__.startswith('int'):
        timestamp = int(timestamp)

    table = schema_utils.get_table_schema('block_timestamps', context=context)
    lower_bound = await toolsql.async_select(
        conn=conn,
        table=table,
        where_lte={'timestamp': timestamp},
        columns=['MAX(block_number)'],
        output_format='cell_or_none',
    )
    upper_bound = await toolsql.async_select(
        conn=conn,
        table=table,
        where_gte={'timestamp': timestamp},
        columns=['MIN(block_number)'],
        output_format='cell_or_none',
    )

    return lower_bound, upper_bound


__all__ = (
    'async_upsert_block_timestamp',
    'async_upsert_block_timestamps',
    'async_delete_block_timestamp',
    'async_delete_block_timestamps',
)



from __future__ import annotations

import math
import typing

import toolcli
import toolstr

from ctc import cli


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': limits_command,
        'help': 'display limits of various datatypes',
        'args': [
            {
                'name': ['--verbose', '-v'],
                'help': 'display all possible bit values',
                'action': 'store_true',
            },
            {
                'name': ['--bits', '-b'],
                'help': 'specify number of bits, either single or as list',
                'nargs': '+',
            },
        ],
        'examples': [
            '',
            '--verbose',
            '--bits 256 512 1024',
        ],
    }


def limits_command(verbose: bool, bits: list[str]) -> None:
    if bits is not None:
        bit_tokens = bits
        bit_tokens = [
            value for token in bit_tokens for value in token.split(',')
        ]
        bit_tokens = [
            value for token in bit_tokens for value in token.split(' ')
        ]
        bit_values = [int(token) for token in bit_tokens]
    elif verbose:
        bit_values = list(range(8, 256 + 8, 8))
    else:
        bit_values = [
            8,
            16,
            32,
            64,
            128,
            192,
            256,
        ]

    styles = cli.get_cli_styles()

    toolstr.print_text_box('int Limits', style=styles['title'])
    rows: typing.List[typing.Sequence[typing.Any]] = []
    for bit_value in bit_values:
        int_min = -(1 << (bit_value - 1))
        int_max = (1 << (bit_value - 1)) - 1
        row = [
            'int' + str(bit_value),
            str(int_min) + '\n ' + str(int_max),
        ]
        row[1] += (
            '\n-10^'
            + toolstr.format(round(math.log10(-int_min)))
            + ' --> 10^'
            + toolstr.format(round(math.log10(int_max)))
        )
        rows.append(row)
    toolstr.print_multiline_table(
        rows,
        column_justify=['right', 'left'],
        column_styles=[
            styles['option'] + ' bold',
            styles['description'],
        ],
        compact=1,
        separate_all_rows=False,
        vertical_justify='top',
    )

    # uint
    print()
    toolstr.print_text_box('uint Limits', style=styles['title'])
    rows = []
    for bit_value in bit_values:
        uint_max = 1 << bit_value
        row = ['uint' + str(bit_value), str(uint_max)]
        row[1] += '\n0 --> 10^' + str(round(math.log10(uint_max)))
        rows.append(row)
    toolstr.print_multiline_table(
        rows,
        column_justify=['right', 'left'],
        column_styles=[
            styles['option'] + ' bold',
            styles['description'],
        ],
        compact=1,
        separate_all_rows=False,
    )

    print()
    floats = [
        ['float16', 1, 5, 10, 3.31, '65500'],
        ['float32', 1, 8, 23, 7.22, '3.40e+38'],
        ['float64', 1, 11, 52, 15.95, '1.80e+308'],
        ['float128', 1, 15, 113, 34.32, '1.19e+4932'],
    ]
    labels = [
        'type',
        'sign\nbits',
        'exp\nbits',
        'mantissa\nbits',
        'decimal\ndigits',
        'max value',
    ]
    column_styles = {label: styles['description'] for label in labels}
    column_styles['type'] = styles['option'] + ' bold'
    toolstr.print_text_box('float Limits', style=styles['title'])
    toolstr.print_table(
        floats,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles=column_styles,  # type: ignore
    )

    print()
    print()
    toolstr.print_text_box('EVM Data', style=styles['title'])

    parameters = [
        ['address', '20 bytes'],
        ['transaction hash', '32 bytes'],
        ['word size', '32 bytes'],
        ['instruction size', '1 byte'],
    ]
    print()
    toolstr.print_table(
        rows=parameters,
        border=styles['comment'],
        column_styles=[styles['metavar'], styles['description']],
    )
    print()
    print()

    rows = [
        ['contract size', '0 bytes', '24576 bytes'],
        ['transaction size', '21000 gas', '30M gas'],
        ['block size', '0 gas', '30M gas'],
    ]
    toolstr.print_table(
        rows,
        labels=['', 'min', 'max'],
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            '': styles['metavar'],
            'min': styles['description'],
            'max': styles['description'],
        },
    )

    print()
    toolstr.print(
        'see '
        + toolstr.add_style('ctc gas -v', styles['option'] + ' bold')
        + ' for EVM gas cost summary',
        style=styles['comment'],
    )



from __future__ import annotations

import toolcli


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': lower_command,
        'help': 'convert string to lower case',
        'args': [
            {'name': 'text', 'help': 'text to convert'},
        ],
        'examples': [
            '0x956F47F50A910163D8BF957Cf5846D573E7f87CA',
        ],
    }


def lower_command(text: str) -> None:
    print(text.lower())



from __future__ import annotations

import toolcli

from ctc import config
from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_address_command,
        'help': """summarize address

for contracts, will display ABI""",
        'args': [
            {'name': 'address', 'help': 'address to get summary of'},
            {
                'name': ['-v', '--verbose'],
                'action': 'store_true',
                'help': 'emit extra output',
            },
            {
                'name': '--raw',
                'action': 'store_true',
                'help': 'emit abi in raw json',
            },
            {
                'name': '--network',
                'metavar': 'NAME_OR_ID',
                'help': 'network name or id to scan address of',
            },
        ],
        'examples': ['0x956f47f50a910163d8bf957cf5846d573e7f87ca'],
    }


async def async_address_command(
    *, address: spec.Address, verbose: bool | int, network: str, raw: bool
) -> None:

    context = config.create_user_input_context(network=network)

    max_width = toolcli.get_n_terminal_cols()

    address = await evm.async_resolve_address(address)

    verbose = 1
    if verbose:
        verbose = 2
    await evm.async_print_address_summary(
        address=address,
        verbose=verbose,
        max_width=max_width,
        raw=raw,
        context=context,
    )



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import uniswap_v3_spec


async def async_get_populated_ticks(
    pool: spec.Address,
    tick_bitmap_index: int,
    *,
    context: spec.Context = None,
) -> tuple[typing.Mapping[str, int], ...]:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'getPopulatedTicksInWord',
        'tick_lens',
    )
    result = await rpc.async_eth_call(
        to_address=uniswap_v3_spec.tick_lens,
        function_abi=function_abi,
        function_parameters=[pool, tick_bitmap_index],
        context=context,
    )
    if not isinstance(result, tuple) or not all(
        isinstance(item, dict) for item in result
    ):
        raise Exception('invalid rpc result')
    return typing.cast(typing.Tuple[typing.Mapping[str, int], ...], result)



from __future__ import annotations

import toolsql

uniswap_v3_pools_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'uniswap_v3_pools': {
            'columns': [
                {'name': 'address', 'type': 'Text', 'primary': True},
                {'name': 'token0', 'type': 'Text'},
                {'name': 'token1', 'type': 'Text'},
                {'name': 'fee', 'type': 'Text'},
            ],
        },
    },
}



from __future__ import annotations

from ctc import spec


def sync_send_websocket(
    request: spec.RpcRequest,
    provider: spec.Provider,
) -> str:
    raise NotImplementedError()


async def async_send_websocket(
    request: spec.RpcRequest,
    provider: spec.Provider,
) -> str:
    raise NotImplementedError()




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec

import numpy as np


# async def async_get_vault_strategies(
#     vault: spec.Address,
# ) -> typing.Sequence[spec.Address]:
#     pass


async def async_get_harvests(
    strategy: spec.Address,
    *,
    include_aprs: bool = True,
    context: spec.Context = None,
) -> spec.DataFrame:
    import polars as pl

    harvests = await evm.async_get_events(
        contract_address=strategy,
        event_name='Harvested',
        include_timestamps=True,
        verbose=False,
        context=context,
    )
    harvests.with_columns(pl.col('arg__profit').apply(int))

    if include_aprs:
        aprs = await async_get_harvest_aprs(
            strategy=strategy, harvests=harvests, context=context
        )
        harvests = harvests.with_columns(pl.Series('apr', aprs))

    return harvests


async def async_get_harvest_aprs(
    strategy: spec.Address,
    *,
    harvests: spec.DataFrame | None = None,
    context: spec.Context = None,
) -> spec.NumpyArray:

    if harvests is None:
        harvests = await async_get_harvests(strategy, context=context)

    durations_task = asyncio.create_task(
        async_get_harvest_durations(harvests, context=context)
    )
    total_debts = await async_get_harvest_total_debts(
        strategy, harvests, context=context
    )
    total_debts = [debt if debt != 0 else float('-inf') for debt in total_debts]
    durations = await durations_task

    aprs = (
        harvests['arg__profit']
        / np.array(total_debts)
        / np.array(durations)
        * 86400
        * 365
    )
    return aprs.to_numpy()


async def async_get_harvest_durations(
    harvests: spec.DataFrame,
    *,
    context: spec.Context = None,
) -> typing.Sequence[int | float]:
    blocks = harvests['block_number'].to_list()
    timestamps = await evm.async_get_block_timestamps(blocks, context=context)
    durations = [float('inf')] + [
        after - before for after, before in zip(timestamps[1:], timestamps[:-1])
    ]
    return durations


async def async_get_harvest_total_debts(
    strategy: spec.Address,
    harvests: spec.DataFrame,
    *,
    context: spec.Context = None,
) -> typing.Sequence[int | float]:

    function_abi: spec.FunctionABI = {
        'inputs': [],
        'name': 'estimatedTotalAssets',
        'outputs': [
            {
                'internalType': 'uint256',
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    }

    total_debts = await rpc.async_batch_eth_call(
        to_address=strategy,
        function_abi=function_abi,
        block_numbers=harvests['block_number'].to_list(),
        context=context,
    )
    return [float('inf')] + total_debts[:-1]




from __future__ import annotations

import os
import shutil
import typing

import toolcli
import toolstr

import ctc
from ctc import spec
from ... import upgrade_utils
from ...upgrade_utils import data_dir_versioning
from ... import config_defaults


def setup_data_dir(
    *,
    styles: toolcli.StyleTheme,
    old_config: typing.Mapping[typing.Any, typing.Any],
    headless: bool,
    default_data_dir: str | None,
    disable_logs: bool,
) -> spec.PartialConfig:

    print()
    print()
    toolstr.print('## Data Root Directory', style=styles['title'])
    print()

    new_data_root = None

    # decide whether to use old data root
    old_data_root: str | None = old_config.get('data_dir')
    if old_data_root is not None and not isinstance(old_data_root, str):
        old_data_root = None
    if isinstance(old_data_root, str):
        old_data_root = os.path.abspath(os.path.expanduser(old_data_root))
        if os.path.isdir(old_data_root):
            prompt = (
                'Continue using data directory ['
                + styles['description']
                + ']'
                + str(old_data_root)
                + '[/'
                + styles['description']
                + ']? '
            )
            if toolcli.input_yes_or_no(
                prompt,
                default='yes',
                style=styles['metavar'],
                headless=headless,
            ):
                new_data_root = old_data_root
            else:
                print(
                    'OK. You can specify whether to keep this old data'
                    ' after a new directory is chosen'
                )

    # enter new data root
    if new_data_root is None:
        if default_data_dir is None:
            default_data_dir = config_defaults.get_default_data_dir()
        new_data_root = toolcli.input_directory_path(
            prompt='Where should ctc store data? (specify a directory path) ',
            default=default_data_dir,
            require_absolute=True,
            must_already_exist=False,
            create_directory=False,
            style=styles['metavar'],
            headless=headless,
        )

        # move old data
        if isinstance(old_data_root, str) and os.path.isdir(old_data_root):
            prompt = 'Move old ctc data to this new location? '
            answer = toolcli.input_yes_or_no(
                prompt,
                default='yes',
                headless=headless,
                style=styles['metavar'],
            )
            if answer:
                if not os.path.isdir(new_data_root):
                    shutil.move(old_data_root, new_data_root)

    # create files and subdirectories and upgrade if need be
    data_dir_versioning.initialize_data_subdirs(
        new_data_root,
        version=upgrade_utils.get_stable_version(ctc.__version__),
    )

    print()
    prompt = 'Do you want to disable ctc logging? '
    if disable_logs:
        default = 'yes'
    else:
        default = 'no'
    disable_logs = toolcli.input_yes_or_no(
        prompt,
        default=default,
        style=styles['metavar'],
        headless=headless,
    )

    return {
        'data_dir': new_data_root,
        'log_rpc_calls': not disable_logs,
        'log_sql_queries': not disable_logs,
    }



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import safe_spec


def parse_safe_signatures(
    signatures: spec.Data,
) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
    """
    reference: https://docs.gnosis-safe.io/contracts/signatures
    """
    as_bytes = evm.to_binary(signatures)

    eip_1271_positions = []
    eip_1271_indices = []
    s = 0
    parsed_signatures = []
    while True:
        signature = as_bytes[:65]
        as_bytes = as_bytes[65:]

        assert len(signature) == 65

        signature_type = signature[-1]
        if 31 > signature_type and signature_type > 26:
            parsed = {
                'type': 'ecdsa',
                'signature': evm.to_hex(signature),
                'r': evm.binary_convert(signature[:32], 'integer'),
                's': evm.binary_convert(signature[32:64], 'integer'),
                'v': evm.binary_convert(signature_type, 'integer'),
            }
        elif signature_type > 30:
            parsed = {
                'type': 'eth_sign',
                'signature': evm.to_hex(signature),
                'r': evm.binary_convert(signature[:32], 'integer'),
                's': evm.binary_convert(signature[32:64], 'integer'),
                'v': evm.binary_convert(signature_type - 4, 'integer'),
            }
        elif signature_type == 0:
            parsed = {
                'type': 'eip1271',
                'signature': evm.to_hex(signature),
                'verifier': evm.to_hex(signature[:32][-20:]),
                'position': evm.binary_convert(signature[32:64], 'integer'),
            }
            position: int = typing.cast(int, parsed['position'])
            eip_1271_positions.append(position)
            eip_1271_indices.append(s)
        elif signature_type == 1:
            parsed = {
                'type': 'prevalidated',
                'signature': evm.to_hex(signature),
                'validator': evm.to_hex(signature[:32][-20:]),
            }
        else:
            raise Exception('unknown signature type: ' + str(signature_type))

        parsed_signatures.append(parsed)
        s += 1

        if len(as_bytes) == 0:
            break
        if len(eip_1271_positions) > 0 and s * 65 >= min(eip_1271_positions):
            break

    return parsed_signatures


def get_safe_transaction_signers(
    *,
    signatures: spec.BinaryData
    | typing.Sequence[typing.Mapping[str, typing.Any]],
    safe_transaction_hash: spec.Data,
    safe_transaction: safe_spec.SafeTransaction,
    chain_id: int,
    safe_address: spec.Address,
    call_data: spec.Data,
    nonce: int,
) -> typing.Sequence[spec.Address]:

    if signatures is None:
        signatures = get_safe_signatures_from_call_data(call_data)

    if safe_transaction_hash is None:
        if (
            call_data is None
            or nonce is None
            or chain_id is None
            or safe_address is None
        ):
            raise Exception(
                'must specify safe_transaction_hash or {call_data, nonce, chain_id, safe_address}'
            )
        safe_transaction = get_safe_transaction_from_call_data(
            call_data=call_data,
            nonce=nonce,
        )
        safe_transaction_hash = hash_safe_transaction(
            safe_transaction=safe_transaction,
            chain_id=chain_id,
            safe_address=safe_address,
        )

    return _get_safe_transaction_signers(
        signatures=signatures,
        safe_transaction_hash=safe_transaction_hash,
    )


def _get_safe_transaction_signers(
    *,
    signatures: spec.BinaryData
    | typing.Sequence[typing.Mapping[str, typing.Any]],
    safe_transaction_hash: spec.Data,
) -> typing.Sequence[spec.Address]:

    if isinstance(signatures, list):
        parsed_signatures: typing.Sequence[
            typing.Mapping[str, typing.Any]
        ] = signatures
    elif isinstance(signatures, (str, bytes)):
        parsed_signatures = parse_safe_signatures(signatures)
    else:
        raise Exception('unknown signatures format: ' + str(type(signatures)))

    signers = []
    for signature in parsed_signatures:

        if signature['type'] in ['ecdsa', 'eth_sign']:
            vrs = signature['v'], signature['r'], signature['s']
            signer = evm.recover_signer_address(
                message_hash=safe_transaction_hash,
                signature=vrs,
            )
        elif signature['type'] == 'eip1271':
            signer = signature['verifier']
        elif signature['type'] == 'prevalidated':
            signer = signature['validator']
        else:
            raise Exception('unknown signature type: ' + str(signature['type']))

        signers.append(signer)

    return signers


def hash_safe_transaction(
    safe_transaction: safe_spec.SafeTransaction,
    *,
    chain_id: int,
    safe_address: spec.Address,
) -> spec.Data:
    """compute the hash to be signed by safe owners"""

    domain = {
        'chain_id': chain_id,
        'verifying_contract': safe_address,
    }
    return evm.hash_eip712_struct(
        struct_data=safe_transaction,
        struct_type=safe_spec.safe_transaction_type,
        domain=domain,
    )


#
# # extracting data from call data
#


def get_safe_transaction_from_call_data(
    call_data: spec.BinaryData,
    nonce: int,
) -> safe_spec.SafeTransaction:

    # decode transaction parameters
    decoded = evm.decode_call_data(
        call_data=call_data,
        function_abi=safe_spec.function_abis['execTransaction'],
    )
    parameters = typing.cast(
        typing.Mapping[str, typing.Any], decoded['named_parameters']
    )

    # create safe transaction from decoded transaction parameters
    safe_transaction: safe_spec.SafeTransaction = {  # type: ignore
        key: parameters[key]
        for key in safe_spec.safe_transaction_keys
        if key != 'nonce'
    }
    safe_transaction['nonce'] = nonce
    return safe_transaction


def get_safe_signatures_from_call_data(
    call_data: spec.BinaryData,
) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
    # decode transaction parameters
    decoded = evm.decode_call_data(
        call_data=call_data,
        function_abi=safe_spec.function_abis['execTransaction'],
    )

    # return signatures
    parameters = typing.cast(
        typing.Mapping[str, typing.Any], decoded['named_parameters']
    )
    raw_signatures = parameters['signatures']
    return parse_safe_signatures(raw_signatures)



"""should move this file to evm.contract_utils"""

from __future__ import annotations

import typing

from ctc import spec
from .. import binary_utils
from .. import block_utils
from .. import transaction_utils
from . import contract_tests


def get_created_address(
    sender: spec.Address,
    nonce: int | None = None,
    *,
    salt: str | None = None,
    init_code: spec.HexData | None = None,
) -> spec.Address:
    """return address created by EVM opcodes CREATE or CREATE2

    see https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1014.md
    see https://ethereum.stackexchange.com/a/761
    """

    if nonce is not None:
        # create
        data: str | bytes = binary_utils.rlp_encode(
            (sender, nonce), str_mode='hex'
        )
    elif salt is not None and init_code is not None:
        # create2
        data = (
            binary_utils.to_hex('0xff', prefix=False)
            + binary_utils.to_hex(sender, prefix=False)
            + binary_utils.to_hex(salt, prefix=False)
            + binary_utils.to_hex(binary_utils.keccak(init_code), prefix=False)
        )
    else:
        raise Exception('specify either {nonce} or {salt, init_code}')

    result = binary_utils.keccak(data, output_format='prefix_hex')
    result = '0x' + result[26:]

    return result


async def async_get_contract_creation_transaction(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    creation_block: int | None = None,
) -> spec.TransactionHash:
    """get hash of contract's creation transaction"""
    from ctc import rpc

    contract_address = contract_address.lower()

    if creation_block is None:
        creation_block = await async_get_contract_creation_block(
            contract_address, context=context
        )
    if creation_block is None:
        raise Exception('could not determine creation block of contract')
    trace_block: spec.TraceList = await rpc.async_trace_block(
        creation_block, context=context
    )
    for item in trace_block:
        if (
            item['type'] == 'create'
            and item['result']['address'] == contract_address  # type: ignore
        ):
            return item['transaction_hash']
    else:
        raise Exception('could not find creation transaction for contract')


async def async_get_contract_deployer(
    contract_address: spec.Address,
    *,
    creation_transaction: spec.TransactionHash | None = None,
    context: spec.Context = None,
) -> spec.Address:
    """get EOA deployer of contract"""
    if creation_transaction is None:
        creation_transaction = await async_get_contract_creation_transaction(
            contract_address, context=context
        )
    tx = await transaction_utils.async_get_transaction(
        creation_transaction, context=context
    )
    return tx['from_address']


async def async_get_contract_creation_block(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    **search_kwargs: typing.Any,
) -> int | None:
    """get block number of when contract was created

    - behavior is undefined for functions that have undergone SELF-DESTRUCT(S)
    - caches result in local database
    """

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='contract_creation_blocks', context=context
    )
    if read_cache:
        from ctc import db

        block = await db.async_query_contract_creation_block(
            address=contract_address,
            context=context,
        )
        if block is not None:
            return block

    block = await _async_get_contract_creation_block_from_node(
        contract_address=contract_address,
        context=context,
        **search_kwargs,
    )

    if write_cache and block is not None:
        await db.async_intake_contract_creation_block(
            contract_address=contract_address,
            block=block,
            context=context,
        )

    return block


async def async_get_contracts_creation_blocks(
    contract_addresses: typing.Sequence[spec.Address],
    *,
    verbose: bool = False,
    context: spec.Context = None,
    **search_kwargs: typing.Any,
) -> typing.Sequence[int | None]:
    """get creation blocks of mutliple contracts

    - behavior is undefined for functions that have undergone SELF-DESTRUCT(S)
    - caches results in local database
    """

    import asyncio

    coroutines = [
        async_get_contract_creation_block(
            contract_address=contract_address,
            context=context,
            verbose=verbose,
            **search_kwargs,
        )
        for contract_address in contract_addresses
    ]
    return await asyncio.gather(*coroutines)


async def _async_get_contract_creation_block_from_node(
    contract_address: spec.Address,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    verbose: bool = True,
    nary: typing.Optional[int] = None,
) -> int | None:
    """get the block where a contract was created

    algorithm: perform a binary search across blocks, check code bytes in each
    """

    from ctc.toolbox import search_utils

    if start_block is None:
        start_block = 0
    if end_block is None:
        end_block = 'latest'
    if start_block == 'latest' or end_block == 'latest':
        latest_block = await block_utils.async_get_latest_block_number(
            context=context
        )
        if start_block == 'latest':
            start_block = latest_block
        if end_block == 'latest':
            end_block = latest_block

    if not isinstance(start_block, int):
        raise Exception('unknown start_block representation')
    if not isinstance(end_block, int):
        raise Exception('unknown end_block representation')

    if verbose:
        print('searching for creation block of ' + contract_address)

    async def async_is_match(index: int) -> bool:
        if verbose:
            print('- trying block:', index)
        return await contract_tests.async_is_contract_address(
            address=contract_address,
            block=index,
            context=context,
        )

    if nary is None:
        result = await search_utils.async_binary_search(
            start_index=start_block,
            end_index=end_block,
            async_is_match=async_is_match,
        )
    else:
        raise NotImplementedError()

    if verbose:
        print('result:', result)

    return result




from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import uniswap_v3_spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class UniswapV3Ticks(TypedDict):
        liquidity_gross: int
        liquidity_net: int
        fee_growth_outside_0_x128: int
        fee_growth_outside_1_x128: int
        tick_cummulative_outside: int
        seconds_per_liquidity_outside_x128: int
        seconds_outside: int
        initialized: bool

    class UniswapV3Observations(TypedDict):
        block_timestamp: int
        tick_cummulative: int
        seconds_per_liquidity_cummulative_x128: int
        initialized: bool


async def async_pool_slot0(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> dict[str, int]:
    function_abi = await uniswap_v3_spec.async_get_function_abi('slot0', 'pool')
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    return {
        'sqrt_price_x96': result[0],
        'tick': result[1],
        'observation_index': result[2],
        'observation_cardinality': result[3],
        'observation_cardinality_next': result[4],
        'fee_protocol': result[5],
        'unlocked': result[6],
    }


async def async_pool_fee_growth_global_0_x128(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'feeGrowthGlobal0X128',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_fee_growth_global_1_x128(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'feeGrowthGlobal1X128',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_protocol_fees(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> tuple[int, int]:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'protocolFees',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if (
        not isinstance(result, tuple)
        or len(result) != 2
        or not isinstance(result[0], int)
        or not isinstance(result[1], int)
    ):
        raise Exception('invalid rpc result')
    return typing.cast(typing.Tuple[int, int], result)


async def async_pool_liquidity(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'liquidity',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_ticks(
    tick: int,
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> UniswapV3Ticks:
    function_abi = await uniswap_v3_spec.async_get_function_abi('ticks', 'pool')
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[tick],
        context=context,
        block_number=block,
    )
    return {
        'liquidity_gross': result[0],
        'liquidity_net': result[1],
        'fee_growth_outside_0_x128': result[2],
        'fee_growth_outside_1_x128': result[3],
        'tick_cummulative_outside': result[4],
        'seconds_per_liquidity_outside_x128': result[5],
        'seconds_outside': result[6],
        'initialized': result[7],
    }


async def async_pool_tick_bitmap(
    word_position: int,
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'tickBitmap',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        context=context,
        block_number=block,
        function_parameters=[word_position],
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_pool_positions(
    key: str,
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> dict[str, int]:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'positions',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[key],
        context=context,
        block_number=block,
    )
    return {
        'liquidity': result[0],
        'fee_growth_inside_0_last_x128': result[1],
        'fee_growth_inside_1_last_x128': result[2],
        'tokens_owed_0': result[3],
        'tokens_owed_1': result[4],
    }


async def async_pool_observations(
    index: int,
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> UniswapV3Observations:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'observations',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[index],
        context=context,
        block_number=block,
    )
    return {
        'block_timestamp': result[0],
        'tick_cummulative': result[1],
        'seconds_per_liquidity_cummulative_x128': result[2],
        'initialized': result[3],
    }




from __future__ import annotations

import toolcli

from .. import llama_tvls


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_chain_command,
        'help': 'output data about a chain tracked by Defi Llama',
        'args': [
            {
                'name': 'chain',
                'help': 'chain to describe',
            },
        ],
        'examples': [
            'Tezos',
        ],
    }


async def async_llama_chain_command(chain: str) -> None:
    await llama_tvls.async_print_historical_chain_tvl(chain)



from __future__ import annotations

import decimal
import typing

from ctc import evm
from ctc import spec

from . import balancer_spec
from . import pool_metadata
from . import pool_state

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    import tooltime


    class BalancerPoolState(TypedDict):
        block: int
        pool_tokens: typing.Sequence[spec.Address]
        pool_fees: typing.Union[int, float]
        pool_weights: typing.Union[
            dict[spec.ContractAddress, int],
            dict[spec.ContractAddress, float],
        ]
        pool_balances: typing.Union[
            dict[spec.Address, int],
            dict[spec.Address, float],
        ]


async def async_summarize_pool_state(
    pool_address: spec.Address,
    block: spec.BlockNumberReference = 'latest',
    *,
    context: spec.Context = None,
) -> BalancerPoolState:
    block = await evm.async_block_number_to_int(block, context=context)

    pool_tokens_coroutine = await pool_metadata.async_get_pool_tokens(
        pool_address=pool_address,
        block=block,
        context=context,
    )
    pool_fees_coroutine = await pool_state.async_get_pool_fees(
        pool_address=pool_address,
        block=block,
        context=context,
    )
    pool_weights_coroutine = await pool_state.async_get_pool_weights(
        pool_address=pool_address,
        block=block,
        context=context,
    )
    pool_balances_coroutine = await pool_state.async_get_pool_balances(
        pool_address=pool_address,
        block=block,
        context=context,
    )

    return {
        'pool_tokens': pool_tokens_coroutine,
        'pool_fees': pool_fees_coroutine,
        'pool_weights': pool_weights_coroutine,
        'pool_balances': pool_balances_coroutine,
        'block': block,
    }


async def async_get_pool_swaps(
    pool_address: typing.Optional[spec.Address] = None,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:
    import polars as pl

    event_abi: spec.EventABI = {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'bytes32',
                'name': 'poolId',
                'type': 'bytes32',
            },
            {
                'indexed': True,
                'internalType': 'contract IERC20',
                'name': 'tokenIn',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'contract IERC20',
                'name': 'tokenOut',
                'type': 'address',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amountIn',
                'type': 'uint256',
            },
            {
                'indexed': False,
                'internalType': 'uint256',
                'name': 'amountOut',
                'type': 'uint256',
            },
        ],
        'name': 'Swap',
        'type': 'event',
    }

    vault = balancer_spec.vault

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    if start_block is None:
        start_block = await evm.async_get_contract_creation_block(
            vault, context=context
        )

    swaps = await evm.async_get_events(
        contract_address=vault,
        event_abi=event_abi,
        start_block=start_block,
        end_block=end_block,
        include_timestamps=include_timestamps,
        verbose=False,
        context=context,
    )

    if pool_address is not None:
        swaps = swaps.filter(
            pl.col('arg__poolId').str.starts_with(pool_address)
        )

    return swaps


def summarize_pool_swaps(
    swaps: spec.DataFrame,
    weights: spec.DataFrame,
    *,
    as_dataframe: bool = True,
) -> typing.Mapping[tuple[str, str], spec.DataFrame]:
    import numpy as np
    import polars as pl

    trade_pairs = set()
    for row in swaps[['arg__tokenIn', 'arg__tokenOut']].rows():
        trade_pairs.add(tuple(row))

    pair_data: typing.MutableMapping[tuple[str, str], spec.DataFrame] = {}
    for token_in, token_out in trade_pairs:
        mask = (swaps['arg__tokenIn'] == token_in) & (
            swaps['arg__tokenOut'] == token_out
        )
        pair_swaps = swaps[mask]
        pair_weights = weights[mask.values]
        in_amounts = pair_swaps['arg__amountIn'].map(
            decimal.Decimal
        ) / decimal.Decimal('1e18')
        out_amounts = pair_swaps['arg__amountOut'].map(
            decimal.Decimal
        ) / decimal.Decimal('1e18')
        price_in_per_out = in_amounts / out_amounts
        price_out_per_in = out_amounts / in_amounts
        cummulative_in_amount = np.cumsum(in_amounts)
        cummulative_out_amount = np.cumsum(out_amounts)
        cummulative_price_in_per_out = (
            cummulative_in_amount / cummulative_out_amount
        )
        cummulative_price_out_per_in = (
            cummulative_out_amount / cummulative_in_amount
        )

        data = {
            'in_amounts': in_amounts,
            'out_amounts': out_amounts,
            'price_in_per_out': price_in_per_out,
            'price_out_per_in': price_out_per_in,
            'cummulative_in_amount': cummulative_in_amount,
            'cummulative_out_amount': cummulative_out_amount,
            'cummulative_price_in_per_out': cummulative_price_in_per_out,
            'cummulative_price_out_per_in': cummulative_price_out_per_in,
        }

        for c, weight_column in enumerate(pair_weights.columns):
            data['weight_' + str(c)] = pair_weights[weight_column].values

        pair_data[(token_in, token_out)] = pl.DataFrame(data)

    return pair_data




from __future__ import annotations

from ctc import spec


def get_stableswap_price(
    x: spec.Number, y: spec.Number, *, A: spec.Number
) -> float:
    k = x + y
    numerator = 1 + 2 * A * (x ** 2) * y / ((k / 2) ** 3)
    denominator = 1 + 2 * A * x * (y ** 2) / ((k / 2) ** 3)
    return float(y / x * numerator / denominator)



from __future__ import annotations

import typing

from ctc import spec

from .. import address_utils
from . import erc20_generic
from . import erc20_normalize
from . import erc20_spec


#
# # total supply
#


async def async_get_erc20_total_supply(
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[int, float] | None:
    """get total supply of ERC20"""

    if block is None:
        block = 'latest'

    result = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='totalSupply',
        block=block,
        context=context,
        **rpc_kwargs,
    )
    if rpc_kwargs.get('convert_reverts_to_none') and result is None:
        return result
    total_supply: int | float = result

    if normalize:
        total_supply = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=total_supply, token=token, context=context
        )

    return total_supply


async def async_get_erc20s_total_supplies(
    tokens: typing.Sequence[spec.ERC20Reference],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get total supplies of ERC20s"""

    if block is None:
        block = 'latest'

    total_supplies = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens, function_name='totalSupply', block=block, context=context, **rpc_kwargs
    )

    if normalize:
        total_supplies = (
            await erc20_normalize.async_normalize_erc20s_quantities(
                tokens=tokens, quantities=total_supplies, context=context
            )
        )

    return total_supplies


async def async_get_erc20_total_supply_by_block(
    token: spec.ERC20Reference,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get historical total supply of ERC20 across multiple blocks"""

    total_supplies = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='totalSupply',
        blocks=blocks,
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        total_supplies = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                token=token,
                quantities=total_supplies,
                context=context,
                blocks=blocks,
            )
        )

    return total_supplies


#
# # balance of
#


async def async_get_erc20_balance(
    wallet: spec.Address,
    token: spec.ERC20Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[int, float] | None:
    """get ERC20 balance"""

    if block is None:
        block = 'latest'

    wallet = await address_utils.async_resolve_address(
        wallet,
        block=block,
        context=context,
    )

    result = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='balanceOf',
        block=block,
        function_parameters=[wallet],
        context=context,
        **rpc_kwargs,
    )
    if result is None:
        if not rpc_kwargs.get('convert_reverts_to_none'):
            raise Exception('invalid result')
        return result
    balance: int | float = result

    if normalize:
        balance = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=balance, token=token, context=context, block=block
        )

    return balance


async def async_get_erc20_balances_of_addresses(
    wallets: typing.Sequence[spec.Address],
    token: spec.ERC20Address,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get ERC20 balance of multiple addresses"""

    from ctc import rpc

    if block is None:
        block = 'latest'

    wallets = await address_utils.async_resolve_addresses(
        wallets,
        block=block,
        context=context,
    )

    balances = await rpc.async_batch_eth_call(
        to_address=token,
        block_number=block,
        function_abi=erc20_spec.erc20_function_abis['balanceOf'],
        function_parameter_list=[[wallet] for wallet in wallets],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        return await erc20_normalize.async_normalize_erc20_quantities(
            quantities=balances, token=token, context=context, block=block
        )
    else:
        return balances


async def async_get_erc20s_balances(
    wallet: spec.Address,
    tokens: typing.Sequence[spec.ERC20Address],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get ERC20 balance of wallet for multiple tokens"""

    from ctc import config

    if block is None:
        block = 'latest'

    wallet = await address_utils.async_resolve_address(
        wallet,
        block=block,
        context=context,
    )

    context = config.update_context(
        context=context,
        merge_provider={'chunk_size': 100},
    )
    balances = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens,
        function_name='balanceOf',
        block=block,
        function_parameters=[wallet],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        balances = await erc20_normalize.async_normalize_erc20s_quantities(
            quantities=balances,
            tokens=tokens,
            context=context,
            block=block,
        )

    return balances


async def async_get_erc20_balance_by_block(
    wallet: spec.Address,
    token: spec.ERC20Reference,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    context: spec.Context = None,
    empty_token: typing.Any = 0,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get historical ERC20 balance over multiple blocks"""

    wallet = await address_utils.async_resolve_address(
        wallet,
        block=blocks[-1],
        context=context,
    )

    balances = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='balanceOf',
        blocks=blocks,
        function_parameters=[wallet],
        context=context,
        empty_token=empty_token,
        **rpc_kwargs,
    )

    if normalize:
        balances = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                quantities=balances,
                token=token,
                context=context,
                blocks=blocks,
            )
        )

    return balances


#
# # allowance
#


async def async_get_erc20_allowance(
    token: spec.ERC20Reference,
    owner: spec.Address,
    *,
    spender: spec.Address,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[int, float] | None:
    """get ERC20 allowance"""

    if block is None:
        block = 'latest'

    owner = await address_utils.async_resolve_address(
        owner,
        block=block,
        context=context,
    )

    result = await erc20_generic.async_erc20_eth_call(
        token=token,
        function_name='allowance',
        block=block,
        function_parameters=[owner, spender],
        context=context,
        **rpc_kwargs,
    )
    if result is None:
        if not rpc_kwargs.get('convert_reverts_to_none'):
            raise Exception('invalid result')
        return result

    allowance: int | float = result

    if normalize:
        allowance = await erc20_normalize.async_normalize_erc20_quantity(
            quantity=allowance, token=token, context=context, block=block
        )

    return allowance


async def async_get_erc20_allowance_by_block(
    token: spec.ERC20Reference,
    owner: spec.Address,
    *,
    spender: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get historical ERC20 allowance over range of blocks"""

    owner = await address_utils.async_resolve_address(
        owner,
        block=blocks[-1],
        context=context,
    )

    allowances = await erc20_generic.async_erc20_eth_call_by_block(
        token=token,
        function_name='allowance',
        blocks=blocks,
        function_parameters=[owner, spender],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        allowances = (
            await erc20_normalize.async_normalize_erc20_quantities_by_block(
                quantities=allowances,
                token=token,
                context=context,
                blocks=blocks,
            )
        )

    return allowances


async def async_get_erc20s_allowances(
    tokens: typing.Sequence[spec.ERC20Reference],
    owner: spec.Address,
    *,
    spender: spec.Address,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Union[typing.Sequence[int], typing.Sequence[float]]:
    """get ERC20 allowance of wallet for multiple tokens"""

    owner = await address_utils.async_resolve_address(
        owner,
        block=block,
        context=context,
    )

    allowances = await erc20_generic.async_erc20s_eth_calls(
        tokens=tokens,
        function_name='allowance',
        block=block,
        function_parameters=[owner, spender],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        allowances = await erc20_normalize.async_normalize_erc20s_quantities(
            quantities=allowances, tokens=tokens, context=context, block=block
        )

    return allowances


async def async_get_erc20_allowances_of_owners(
    token: spec.ERC20Reference,
    owners: typing.Sequence[spec.Address],
    *,
    spender: spec.Address,
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[int | float]:
    """get ERC20 allowance of multiple owners"""

    from ctc import rpc

    owners = await address_utils.async_resolve_addresses(
        owners,
        block=block,
        context=context,
    )

    allowances = await rpc.async_batch_eth_call(
        to_address=token,
        block_number=block,
        function_abi=erc20_spec.erc20_function_abis['allowance'],
        function_parameter_list=[[owner, spender] for owner in owners],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        return await erc20_normalize.async_normalize_erc20_quantities(
            quantities=allowances, token=token, context=context, block=block
        )
    else:
        return allowances


async def async_get_erc20_allowances_of_spenders(
    token: spec.ERC20Reference,
    owner: spec.Address,
    *,
    spenders: typing.Sequence[spec.Address],
    block: spec.BlockNumberReference | None = None,
    normalize: bool = True,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[int | float]:
    """get ERC20 allowance of multiple spenders"""

    from ctc import rpc

    owner = await address_utils.async_resolve_address(
        owner,
        block=block,
        context=context,
    )

    allowances = await rpc.async_batch_eth_call(
        to_address=token,
        block_number=block,
        function_abi=erc20_spec.erc20_function_abis['allowance'],
        function_parameter_list=[[owner, spender] for spender in spenders],
        context=context,
        **rpc_kwargs,
    )

    if normalize:
        return await erc20_normalize.async_normalize_erc20_quantities(
            quantities=allowances, token=token, context=context, block=block
        )
    else:
        return allowances




from __future__ import annotations

import typing

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_trace_transaction(
    transaction_hash: str,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_transaction(
        transaction_hash=transaction_hash
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_transaction(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_trace_replay_transaction(
    transaction_hash: str,
    trace_type: typing.Sequence[spec.TraceOutputType],
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_replay_transaction(
        transaction_hash=transaction_hash, trace_type=trace_type
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_replay_transaction(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_trace_raw_transaction(
    call_data: str,
    trace_type: typing.Sequence[spec.TraceOutputType],
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    raise NotImplementedError()


#
# # calls
#


async def async_trace_call(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
    trace_type: typing.Sequence[spec.TraceOutputType],
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    block_number: spec.BlockNumberReference | None = None,
) -> spec.RpcSingularResponse:

    request = rpc_constructors.construct_trace_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
        trace_type=trace_type,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_call(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


async def async_trace_call_many(
    calls: typing.Sequence[typing.Mapping[str, typing.Any]],
    trace_type: typing.Sequence[spec.TraceOutputType] | None,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_call_many(
        calls=calls,
        trace_type=trace_type,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_call_many(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
    )


#
# # trace ranges
#


async def async_trace_get(
    transaction_hash: spec.TransactionHash,
    trace_indices: typing.Sequence[int],
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_get(
        transaction_hash=transaction_hash,
        trace_indices=trace_indices,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_get(response=response)


async def async_trace_filter(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    from_addresses: spec.Address | None = None,
    to_addresses: spec.Address | None = None,
    after: int | None = None,
    count: int | None = None,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_filter(
        start_block=start_block,
        end_block=end_block,
        from_addresses=from_addresses,
        to_addresses=to_addresses,
        after=after,
        count=count,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_trace_filter(response=response)


async def async_trace_block(
    block_number: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    raw_output: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_block(
        block_number=block_number,
    )
    response = await rpc_request.async_send(
        request, context=context, raw_output=raw_output
    )
    if raw_output:
        return response
    else:
        return rpc_digestors.digest_trace_block(
            response=response,
            decode_response=decode_response,
            snake_case_response=snake_case_response,
        )


async def async_trace_replay_block_transactions(
    block_number: spec.BlockNumberReference,
    trace_type: typing.Sequence[spec.TraceOutputType] | None,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    raw_output: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_trace_replay_block_transactions(
        block_number=block_number,
        trace_type=trace_type,
    )
    response = await rpc_request.async_send(
        request,
        context=context,
        raw_output=raw_output,
    )
    if raw_output:
        return response
    else:
        return rpc_digestors.digest_trace_replay_block_transactions(
            response=response,
            decode_response=decode_response,
            snake_case_response=snake_case_response,
        )


#
# # debug traces
#


async def async_debug_trace_call(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
    trace_type: spec.TraceOutputType,
    context: spec.Context = None,
    snake_case_response: bool = True,
    block_number: spec.BlockNumberReference | None = None,
) -> spec.RpcSingularResponse:

    request = rpc_constructors.construct_debug_trace_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
        trace_type=trace_type,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_debug_trace_call(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_debug_trace_call_many(
    calls: typing.Sequence[typing.Mapping[str, typing.Any]],
    trace_type: spec.TraceOutputType | None,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_debug_trace_call_many(
        calls=calls,
        trace_type=trace_type,
        block_number=block_number,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_debug_trace_call_many(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_debug_trace_transaction(
    transaction_hash: str,
    *,
    context: spec.Context = None,
    trace_type: spec.TraceOutputType | None = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    request = rpc_constructors.construct_debug_trace_transaction(
        transaction_hash=transaction_hash,
        trace_type=trace_type,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_debug_trace_transaction(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_debug_trace_block_by_number(
    block_number: spec.BlockNumberReference,
    *,
    context: spec.Context = None,
    trace_type: spec.TraceOutputType | None = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_debug_trace_block_by_number(
        block_number=block_number,
        trace_type=trace_type,
    )
    response = await rpc_request.async_send(
        request,
        context=context,
    )
    return rpc_digestors.digest_debug_trace_block_by_number(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_debug_trace_block_by_hash(
    block_hash: str,
    *,
    context: spec.Context = None,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_debug_trace_block_by_hash(
        block_hash=block_hash,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_debug_trace_block_by_hash(
        response=response,
        snake_case_response=snake_case_response,
    )




from __future__ import annotations

import re
import typing

from ctc import evm


if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')


def decode_response(
    response: dict[str, typing.Union[int, str]],
    quantities: typing.Optional[list[str]] = None,
) -> dict[str, typing.Any]:
    if quantities is None:
        quantities = []
    decoded = {}
    for key, value in response.items():
        if key in quantities:
            value = evm.binary_convert(value, 'integer')
        decoded[key] = value
    return decoded


def keys_to_snake_case(map: dict[str, T]) -> dict[str, T]:
    """

    not for general usage beyond the keys that appear in the EVM rpc spec

    adapted from https://stackoverflow.com/a/1176023
    """
    return {camel_case_to_snake_case(key): value for key, value in map.items()}


def camel_case_to_snake_case(text: str) -> str:
    """

    not for general usage beyond the keys that appear in the EVM rpc spec

    adapted from https://stackoverflow.com/a/1176023
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', text).lower()



from .aave_interest_rates import *
from .aave_lending_pool import *
from .aave_oracle import *
from .aave_pool_tokens import *
from .aave_spec import *
from .aave_summaries import *
from .aave_rewards import *



from __future__ import annotations

import asyncio
import typing

from ctc import rpc
from ctc import spec

from .. import fuse_queries
from . import lens_abis
from . import lens_spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class _ReturnPoolSummaryOrError(TypedDict):
        summary: typing.Optional[lens_spec.ReturnPoolSummary]
        error: typing.Optional[str]


async def async_get_public_pools_with_data(
    *,
    lens_address: spec.Address | None = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> lens_spec.PublicPoolsWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)

    # on-chain implementation reverts, python implementation instead
    all_pools = await fuse_queries.async_get_all_pools(
        block=block, context=context
    )
    comptrollers = [pool[2] for pool in all_pools]
    coroutines = [
        _async_get_pool_summary_or_error(
            comptroller, block=block, context=context
        )
        for comptroller in comptrollers
    ]
    results = await asyncio.gather(*coroutines)
    summaries = [result['summary'] for result in results]
    errors = [result['error'] is not None for result in results]

    return {
        'public_pools': [
            lens_spec.fuse_pool_to_dict(pool) for pool in all_pools
        ],
        'data': summaries,
        'errored': errors,
    }


async def _async_get_pool_summary_or_error(
    comptroller: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> _ReturnPoolSummaryOrError:
    try:
        summary = await async_get_pool_summary(
            comptroller=comptroller,
            lens_address=lens_address,
            block=block,
            context=context,
        )
        return {'summary': summary, 'error': None}
    except spec.RpcException as e:
        return {'summary': None, 'error': e.args[0]}


async def async_get_public_pools_by_verification_with_data(
    *,
    whitelisted_admin: bool,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.PublicPoolsWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)

    # on-chain implementation reverts, python implementation instead
    all_pools = await fuse_queries.async_get_all_pools(
        block=block, context=context
    )
    comptrollers = [pool[2] for pool in all_pools]
    coroutines = [
        _async_get_pool_summary_or_error(
            comptroller, block=block, context=context
        )
        for comptroller in comptrollers
    ]
    results = await asyncio.gather(*coroutines)

    filtered_pools = []
    filtered_results = []
    for pool, result in zip(all_pools, results):
        if (
            result['summary'] is not None
            and result['summary']['whitelisted_admin'] == whitelisted_admin
        ):
            filtered_pools.append(pool)
            filtered_results.append(result)
    summaries = [result['summary'] for result in filtered_results]
    errors = [result['error'] is not None for result in filtered_results]

    return {
        # 'indexes': result[0],
        'public_pools': [
            lens_spec.fuse_pool_to_dict(pool) for pool in filtered_pools
        ],
        'data': summaries,
        'errored': errors,
    }


async def async_get_pools_by_account_with_data(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolsBySupplierWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolsByAccountWithData')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )
    return {
        'indices': result[0],
        'account_pools': [
            lens_spec.fuse_pool_to_dict(pool) for pool in result[1]
        ],
        'data': [lens_spec.fuse_pool_data_to_dict(pool) for pool in result[2]],
        'errored': result[3],
    }


async def async_get_pool_summary(
    comptroller: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolSummary:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolSummary')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[comptroller],
        block_number=block,
        context=context,
    )
    return lens_spec.return_pool_summary_to_dict(result)


async def async_get_pool_assets_with_data(
    comptroller: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> list[lens_spec.FusePoolAsset]:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolAssetsWithData')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[comptroller],
        block_number=block,
        context=context,
    )
    return [lens_spec.fuse_pool_asset_to_dict(item) for item in result]


async def async_get_public_pool_users_with_data(
    *,
    max_health: int = int(1e36),
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPublicPoolUsersWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)

    # on-chain implementation reverts, python implementation instead
    # kwargs = {'block': block, 'provider': provider}
    max_health = int(1e36)

    all_pools = await fuse_queries.async_get_all_pools(
        block=block, context=context
    )
    comptrollers = [pool[2] for pool in all_pools]
    coroutines = [
        _async_get_pool_users_with_data_or_error(
            comptroller=comptroller,
            max_health=max_health,
            block=block,
            context=context,
        )
        for comptroller in comptrollers
    ]
    result = await asyncio.gather(*coroutines)

    users = []
    close_factors = []
    liquidation_incentives = []
    errored = []
    for subresult in result:
        if subresult['error'] is None:
            users.append(subresult['pool_users']['users'])
            close_factors.append(subresult['pool_users']['close_factor'])
            liquidation_incentives.append(
                subresult['pool_users']['liquidation_incentive']
            )
            errored.append(False)
        else:
            users.append(None)
            close_factors.append(None)
            liquidation_incentives.append(None)
            errored.append(True)

    return {
        'comptrollers': comptrollers,
        'users': users,
        'close_factors': close_factors,
        'liquidation_incentives': liquidation_incentives,
        'errored': errored,
    }


async def _async_get_pool_users_with_data_or_error(
    **kwargs: typing.Any,
) -> dict[str, typing.Any]:
    try:
        summary = await async_get_pool_users_with_data(**kwargs)
        return {'pool_users': summary, 'error': None}
    except spec.RpcException as e:
        return {'pool_users': None, 'error': e.args[0]}


async def async_get_pool_users_with_data(
    comptroller: spec.Address,
    *,
    max_health: int = int(1e36),
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolUsersWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi(
        'getPoolUsersWithData',
        parameter_types=('address', 'uint256'),
    )
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[comptroller, max_health],
        block_number=block,
        context=context,
    )

    return lens_spec.return_pool_users_with_data_to_dict(result)


async def async_get_pools_users_with_data(
    comptrollers: spec.Address,
    *,
    max_health: int = int(1e36),
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolsUsersWithData:
    """renamed to prevent name collision with async_get_pool_users_with_data"""

    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi(
        'getPoolUsersWithData',
        parameter_types=('address[]', 'uint256'),
    )
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[comptrollers, max_health],
        block_number=block,
        context=context,
    )

    return {
        'users': [
            [lens_spec.fuse_pool_user_to_dict(user) for user in pool_users]
            for pool_users in result[0]
        ],
        'close_factors': result[1],
        'liquidation_incentives': result[2],
        'errored': result[3],
    }


async def async_get_pools_by_supplier(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolsBySupplier:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolsBySupplier')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )
    return {
        'indices': result[0],
        'account_pools': result[1],
    }


async def async_get_pools_by_supplier_with_data(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnPoolsBySupplierWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolsBySupplierWithData')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )
    return {
        'indices': result[0],
        'account_pools': [
            lens_spec.fuse_pool_to_dict(pool) for pool in result[1]
        ],
        'data': [lens_spec.fuse_pool_data_to_dict(pool) for pool in result[2]],
        'errored': result[3],
    }


async def async_get_user_summary(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.UserSummary:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getUserSummary')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )

    return {
        'supply_balance': result[0],
        'borrow_balance': result[1],
        'error': result[2],
    }


async def async_get_pool_user_summary(
    comptroller: spec.Address,
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.PoolUserSummary:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getPoolUserSummary')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[comptroller, account],
        block_number=block,
        context=context,
    )
    return {
        'supply_balance': result[0],
        'borrow_balance': result[1],
    }


async def async_get_whitelisted_pools_by_account(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> typing.Sequence[lens_spec.FusePool]:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi('getWhitelistedPoolsByAccount')
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )
    output = result[1]
    return typing.cast(typing.Sequence[lens_spec.FusePool], output)


async def async_get_whitelisted_pools_by_account_with_data(
    account: spec.Address,
    *,
    lens_address: spec.Address | None = None,
    context: spec.Context = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> lens_spec.ReturnWhitelistedPoolsByAccountWithData:
    if lens_address is None:
        lens_address = lens_spec.get_lens_address('primary', context=context)
    function_abi = lens_abis.get_function_abi(
        'getWhitelistedPoolsByAccountWithData'
    )
    result = await rpc.async_eth_call(
        to_address=lens_address,
        function_abi=function_abi,
        function_parameters=[account],
        block_number=block,
        context=context,
    )

    output: lens_spec.ReturnWhitelistedPoolsByAccountWithData = {
        'indices': result[0],
        'account_pools': result[1],
        'data': result[2],
        'errored': result[3],
    }

    output['account_pools'] = [
        lens_spec.fuse_pool_to_dict(item)
        for item in typing.cast(
            typing.List[typing.Any], output['account_pools']
        )
    ]

    output['data'] = [
        lens_spec.fuse_pool_data_to_dict(item)
        for item in typing.cast(typing.List[typing.Any], output['data'])
    ]

    return output




"""ctc is a tool for collecting and processing historical EVM data"""

from .evm import *


__version__ = '0.3.8'


def _clean_package_imports() -> None:
    """remove deep nested modules from ctc namespace"""

    import sys

    ctc = sys.modules['ctc']
    moduletype = type(ctc)
    delattr(ctc, 'annotations')
    for key, value in tuple(vars(ctc).items()):
        if isinstance(value, moduletype):
            name = value.__name__
            if not name.startswith('ctc') or name.count('.') > 1:
                delattr(ctc, key)


_clean_package_imports()



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

from . import aave_spec

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_deposits(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    if end_block is None:
        end_block = 'latest'

    aave_lending_pool = aave_spec.get_aave_address(
        name='LendingPool',
        context=context,
    )
    events = await evm.async_get_events(
        contract_address=aave_lending_pool,
        event_name='Deposit',
        start_block=start_block,
        end_block=end_block,
        include_timestamps=include_timestamps,
        verbose=False,
        context=context,
    )

    return events


async def async_get_withdrawals(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    include_timestamps: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:

    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )

    if end_block is None:
        end_block = 'latest'

    aave_lending_pool = aave_spec.get_aave_address(
        name='LendingPool',
        context=context,
    )
    events = await evm.async_get_events(
        contract_address=aave_lending_pool,
        event_name='Withdraw',
        start_block=start_block,
        include_timestamps=include_timestamps,
        end_block=end_block,
        verbose=False,
        context=context,
    )

    return events



from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import TypeGuard

    from .. import typedefs


def is_polars_dataframe(item: typing.Any) -> TypeGuard[spec.DataFrame]:
    """return whether input is a polars DataFrame"""
    item_type = type(item)
    return (
        item_type.__name__ == 'DataFrame'
        and item_type.__module__ == 'polars.internals.dataframe.frame'
    )


def is_polars_series(item: typing.Any) -> TypeGuard[spec.DataFrame]:
    """return whether input is a polars DataFrame"""
    item_type = type(item)
    return item_type.__name__ == 'Series' and item_type.__module__.startswith(
        'polars'
    )


def is_int(value: typing.Any) -> TypeGuard['typedefs.Integer']:
    """return whether input is a python int or numpy int"""
    return isinstance(value, int) or type(value).__name__ in (
        'int8',
        'int16',
        'int32',
        'int64',
    )


def is_float(value: typing.Any) -> TypeGuard['typedefs.Float']:
    """return whether input is a python float or numpy float"""
    return isinstance(value, float) or type(value).__name__ in (
        'float16',
        'float32',
        'float64',
        'float128',
    )


def is_number(value: typing.Any) -> TypeGuard['typedefs.Number']:
    """return whether input is any type of int or float"""
    return is_int(value) or is_float(value)




from __future__ import annotations

import typing

import toolstr
import toolcli

import ctc
from . import setup_io
from .stages import alias_setup
from .stages import data_dir_setup
from .stages import db_setup
from .stages import network_setup
from .stages import cli_setup


styles: toolcli.StyleTheme = {
    'title': '#ce93f9 bold',
    'description': '#b9f29f bold',
    'metavar': '#8be9fd',
    'content': '#f1fa8c',
    'option': '#64aaaa',
    'comment': '#6272a4',
}


async def async_setup_ctc(
    *,
    headless: bool = False,
    ignore_old_config: bool = False,
    skip_networking: bool = False,
    rpc_url: str | None = None,
    rpc_chain_id: int | None = None,
    data_dir: str | None = None,
    disable_logs: bool = False,
    skip_db: bool = False,
    overwrite: bool = True,
    skip_aliases: bool = False,
) -> None:

    # print intro
    toolstr.print('# Setting up ctc...', style=styles['title'])
    print()
    toolstr.print(
        'Running setup process for ctc '
        + toolstr.add_style(ctc.__version__, styles['description'] + ' bold')
    )
    if not ctc.__version__.split('.')[-1].isnumeric():
        print('(this is not a stable release, use pypi for stable releases)')
    print()
    print('Each step is optional')
    print('- by default, setup process will leave existing settings unchanged')
    print('- setup can be rerun multiple times idempotently')
    print()
    toolstr.print(
        'Can skip options by pressing enter at each prompt', style='bold'
    )

    # load old config data for passing to each option
    if ignore_old_config:
        old_config: typing.Mapping[typing.Any, typing.Any] = {}
    else:
        old_config = setup_io.load_old_config(convert_to_latest=True)
    setup_io.setup_config_path()

    # collect new config file data
    network_data = await network_setup.async_setup_networks(
        skip_networking=skip_networking,
        rpc_url=rpc_url,
        rpc_chain_id=rpc_chain_id,
        old_config=old_config,
        styles=styles,
        headless=headless,
    )
    data_dir_data = data_dir_setup.setup_data_dir(
        old_config=old_config,
        styles=styles,
        headless=headless,
        default_data_dir=data_dir,
        disable_logs=disable_logs,
    )
    cli_data = cli_setup.setup_cli(
        old_config=old_config,
        styles=styles,
        headless=headless,
    )
    if not skip_db:
        db_data = db_setup.setup_dbs(
            data_dir=data_dir_data['data_dir'],
            network_data=network_data,
            styles=styles,
            headless=headless,
            old_db_configs=old_config.get('db_configs'),
        )
    else:
        from ctc.config import config_defaults

        db_data = {
            'db_configs': config_defaults.get_default_db_configs(
                data_dir=data_dir_data['data_dir']
            ),
        }

    # create new config file if need be
    setup_io.write_new_config(
        network_data=network_data,
        db_data=db_data,
        data_dir_data=data_dir_data,
        cli_data=cli_data,
        styles=styles,
        headless=headless,
        overwrite=overwrite,
    )

    # populate db
    if not skip_db:
        await db_setup.async_populate_db_tables(
            db_config=db_data['db_configs']['main'],
            styles=styles,
        )

    # setup aliases
    alias_setup.add_cli_aliases(
        styles=styles,
        headless=headless,
        skip_aliases=skip_aliases,
    )

    # finalize
    print()
    print()
    toolstr.print('## Final Steps', style=styles['title'])
    print()
    print('ctc setup complete')



from __future__ import annotations

from ... import query_utils
from . import blocks_statements


async_query_block = query_utils.wrap_selector_with_connection(
    blocks_statements.async_select_block,
    'blocks',
)

async_query_blocks = query_utils.wrap_selector_with_connection(
    blocks_statements.async_select_blocks,
    'blocks',
)



from __future__ import annotations

import sys
import typing

import ctc
from ctc import spec
from .. import config_defaults
from . import version_utils


def get_config_upgrade_functions() -> (
    typing.Mapping[
        str,
        typing.Callable[
            [typing.MutableMapping[typing.Any, typing.Any]],
            typing.MutableMapping[typing.Any, typing.Any],
        ],
    ]
):
    return {
        '0.2.': upgrade__0_2_0__to__0_3_0,
        '0.3.0': upgrade__0_3_0__to__0_3_1,
        '0.3.1': upgrade__0_3_1__to__0_3_2,
        '0.3.2': upgrade__0_3_2__to__0_3_3,
        '0.3.3': upgrade__0_3_3__to__0_3_4,
        '0.3.4': upgrade__0_3_4__to__0_3_5,
        '0.3.5': upgrade__0_3_5__to__0_3_6,
    }


def upgrade_config(
    old_config: typing.Mapping[typing.Any, typing.Any]
) -> typing.MutableMapping[str, typing.Any]:
    """upgrade config to latest version as much as possible"""

    import string

    # detect version
    version = old_config.get('config_spec_version')
    if version is None:
        version = old_config.get('version')

    # perform upgrade
    if not isinstance(version, str):
        print(
            'old_config has unknown version, using default config',
            file=sys.stderr,
        )
        return dict(config_defaults.get_default_config())

    new_config: typing.MutableMapping[typing.Any, typing.Any] = dict(old_config)
    config_version = version
    current_version_clean = ctc.__version__.rstrip(string.ascii_letters)
    upgrade_functions = get_config_upgrade_functions()

    # check that current config version is recognized
    if not config_version.startswith(current_version_clean) and not any(
        config_version.startswith(upgrade_version)
        for upgrade_version in upgrade_functions.keys()
    ):
        raise Exception('invalid config version')

    # update config from old version using upgrade path
    for from_version, upgrade_function in upgrade_functions.items():
        if config_version.startswith(from_version):
            new_config = upgrade_function(new_config)
            config_version = new_config['config_spec_version']

    new_config_stable = version_utils.get_stable_version(
        new_config['config_spec_version']
    )
    ctc_stable = version_utils.get_stable_version(ctc.__version__)
    if (
        new_config_stable == ctc_stable
        and new_config['config_spec_version'] != ctc.__version__
    ):
        new_config['config_spec_version'] = ctc.__version__

    return new_config

    # # ? perform validation


def upgrade__0_2_0__to__0_3_0(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    network_defaults = upgraded.pop('network_defaults', {})
    upgraded['default_network'] = network_defaults.get('default_network')
    upgraded['default_providers'] = network_defaults.get(
        'default_providers', {}
    )

    # add default networks
    upgraded['networks'] = dict(upgraded.get('networks', {}))
    default_networks = config_defaults.get_default_networks_metadata()
    for chain_id, network_metadata in default_networks.items():
        name = network_metadata['name']
        if name not in upgraded['networks']:
            upgraded['networks'][name] = network_metadata

    # convert to integer network references
    chain_ids_by_network_name = {
        network_metadata['name']: network_metadata['chain_id']
        for network_name, network_metadata in upgraded['networks'].items()
    }
    upgraded['networks'] = {
        chain_ids_by_network_name[network_name]: network_metadata
        for network_name, network_metadata in upgraded['networks'].items()
    }

    if upgraded['default_network'] not in chain_ids_by_network_name:
        raise spec.ConfigUpgradeError(
            'unknown chain_id for network ' + str(upgraded['default_network'])
        )
    upgraded['default_network'] = chain_ids_by_network_name[
        upgraded['default_network']
    ]
    upgraded['default_providers'] = {
        chain_ids_by_network_name[network_name]: provider_name
        for network_name, provider_name in upgraded['default_providers'].items()
    }

    # set provider network references to chain_id's instead of network names
    new_providers = {}
    for provider_name, provider in upgraded['providers'].items():
        new_providers[provider_name] = dict(provider)
        old_network = new_providers[provider_name].get('network')
        if isinstance(old_network, str):
            new_providers[provider_name]['network'] = chain_ids_by_network_name[
                old_network
            ]
    upgraded['providers'] = new_providers

    # set db config
    default_db_config = config_defaults.get_default_db_config(
        data_dir=old_config['data_dir']
    )
    upgraded['db_configs'] = {'main': default_db_config}

    upgraded['log_rpc_calls'] = True
    upgraded['log_sql_queries'] = True

    # set new version
    if 'version' in upgraded:
        del upgraded['version']
    upgraded['config_spec_version'] = '0.3.0'

    return upgraded


def upgrade__0_3_0__to__0_3_1(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    if 'cli_color_theme' not in upgraded:
        upgraded[
            'cli_color_theme'
        ] = config_defaults.get_default_cli_color_theme()
    if 'cli_chart_charset' not in upgraded:
        upgraded[
            'cli_chart_charset'
        ] = config_defaults.get_default_cli_chart_charset()
    upgraded['config_spec_version'] = '0.3.1'

    return upgraded


def upgrade__0_3_1__to__0_3_2(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    default_config: typing.Mapping[
        str, typing.Any
    ] = config_defaults.get_default_config()
    for key in [
        'context_cache_rules',
    ]:
        if key not in upgraded:
            upgraded[key] = default_config[key]

    if old_config.get('default_network') is None:
        old_config['default_network'] = 1

    upgraded['config_spec_version'] = '0.3.2'

    return upgraded


def upgrade__0_3_2__to__0_3_3(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    upgraded['config_spec_version'] = '0.3.3'
    return upgraded


def upgrade__0_3_3__to__0_3_4(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    upgraded['config_spec_version'] = '0.3.4'
    for provider in upgraded['providers'].values():
        provider.setdefault('disable_batch_requests', False)
    for network in upgraded['networks'].values():
        if network['name'] == 'mainnet':
            network['name'] = 'ethereum'
    return upgraded


def upgrade__0_3_4__to__0_3_5(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    upgraded['config_spec_version'] = '0.3.5'
    return upgraded


def upgrade__0_3_5__to__0_3_6(
    old_config: typing.MutableMapping[typing.Any, typing.Any]
) -> typing.MutableMapping[typing.Any, typing.Any]:
    upgraded = dict(old_config)
    upgraded['config_spec_version'] = '0.3.6'
    return upgraded


def omit_extra_version_data(version: str) -> str:
    for substr in ['a', 'b', 'rc']:
        if substr in version:
            index = version.index(substr)
            version = version[:index]
    return version




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

from . import pool_summary


# async def async_plot_lbp_summary(
#     *,
#     swaps: spec.DataFrame,
#     weights: spec.DataFrame,
#     pool_name: str,
#     pool_tokens: typing.Sequence[spec.Address],
#     pool_address: spec.Address,
#     price_range: typing.Sequence[int | float] | None = None,
#     premium_range: typing.Sequence[int | float] | None = None,
#     oracle_data: spec.DataFrame | None = None,
#     context: spec.Context = None,
# ) -> None:

#     import polars as pl

#     summary = pool_summary.summarize_pool_swaps(swaps=swaps, weights=weights)

#     for pair in summary.keys():

#         blocks = summary[pair]['block_number'].to_list()

#         token_0, token_1 = await evm.async_get_erc20s_symbols(
#             pool_tokens, context=context
#         )
#         in_address, out_address = pair
#         if in_address == pool_tokens[0]:
#             in_token = token_0
#             out_token = token_1
#             out_weights = summary[pair]['weight_1']
#             # in_weights = summary[pair]['weight_0']
#         else:
#             out_token = token_0
#             in_token = token_1
#             out_weights = summary[pair]['weight_0']
#             # in_weights = summary[pair]['weight_1']

#         title = pool_name + ', in=' + in_token + ', out=' + out_token

#         if oracle_data is not None:
#             min_block = swaps['block_number'][0]
#             max_block = swaps['block_number'][-1]
#             oracle_data = oracle_data.filter(
#                 (pl.col('block_number') >= min_block)
#                 & (pl.col('block_number') <= max_block)
#             )
#             if oracle_data is None:
#                 raise Exception('invalid oracle_data')
#             ys = [
#                 {
#                     'y': oracle_data.values,
#                     'x': oracle_data['block_number'].to_list(),
#                     'y_kwargs': {'label': 'oracle'},
#                 },
#             ]

#             if oracle_data is None:
#                 raise Exception('invalid oracle_data')
#             premium = []
#             for index, row in summary[pair].iterrows():
#                 block_number = index[0]
#                 oracle_price = oracle_data[block_number]
#                 actual_price = row['price_out_per_in']
#                 premium.append(float(actual_price) / oracle_price - 1)
#         else:
#             ys = None
#             premium = []

#         y_kwargs = {'marker': '.', 'markersize': 10, 'linewidth': 1}
#         plot_data = {
#             'subplot_height': 5,
#             'common': {
#                 'x': blocks,
#                 'name_position': 'ylabel',
#                 'tickgrid': True,
#                 'y_kwargs': y_kwargs,
#             },
#             'plots': {
#                 'price': {
#                     'title': title,
#                     'name': 'price\n(' + out_token + ' / ' + in_token + ')',
#                     'y': summary[pair]['price_out_per_in'],
#                     'y_kwargs': dict(label='actual', **y_kwargs),
#                     'ys': ys,
#                     'ylim': price_range,
#                     # 'legend_kwargs': {'loc': 'lower left'},
#                 },
#                 'premium': {
#                     'name': 'price premium',
#                     'y': premium,
#                     'ylim': premium_range,
#                 },
#                 'weight': {
#                     'name': out_token + ' Weights',
#                     'y': out_weights,
#                 },
#                 'token0 Sold': {
#                     'name': in_token + ' Amount',
#                     'y': summary[pair]['in_amounts'],
#                 },
#                 'Cummulative token0 Bought': {
#                     'name': 'Cummulative\n' + in_token + ' Bought',
#                     'y': summary[pair]['cummulative_in_amount'],
#                 },
#                 'Cummulative token1 Sold': {
#                     'name': 'Cummulative\n' + out_token + 'Sold',
#                     'y': summary[pair]['cummulative_out_amount'],
#                 },
#                 'Volume Weighted Price': {
#                     'name': 'Cummulative\nVolume Weighted\nPrice',
#                     'y': summary[pair]['cummulative_price_out_per_in'],
#                     'xlabel': 'block',
#                     'ylim': price_range,
#                 },
#             },
#         }

#         import toolplot  # type: ignore

#         toolplot.plot_subplots(plot_data)




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:

    from typing_extensions import TypedDict

    import toolsql

    class CoingeckoToken(TypedDict):
        id: str
        symbol: str
        name: str
        market_cap_rank: int


coingecko_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'coingecko_tokens': {
            'columns': [
                {'name': 'id', 'type': 'Text', 'primary': True},
                {
                    'name': 'symbol',
                    'type': 'Text',
                    'index': True,
                },
                {'name': 'name', 'type': 'Text', 'index': True},
                {'name': 'market_cap_rank', 'type': 'Integer', 'index': True},
            ],
        },
    },
}
